
import google.protobuf
from packaging.version import Version
if Version(google.protobuf.__version__).major >= 5:
  # -*- coding: utf-8 -*-
  # Generated by the protocol buffer compiler.  DO NOT EDIT!
  # source: databricks_filesystem_service.proto
  # Protobuf Python Version: 5.26.0
  """Generated protocol buffer code."""
  from google.protobuf import descriptor as _descriptor
  from google.protobuf import descriptor_pool as _descriptor_pool
  from google.protobuf import symbol_database as _symbol_database
  from google.protobuf.internal import builder as _builder
  # @@protoc_insertion_point(imports)

  _sym_db = _symbol_database.Default()


  from . import databricks_pb2 as databricks__pb2
  from .scalapb import scalapb_pb2 as scalapb_dot_scalapb__pb2


  DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n#databricks_filesystem_service.proto\x12\x11mlflow.filesystem\x1a\x10\x64\x61tabricks.proto\x1a\x15scalapb/scalapb.proto\")\n\nHttpHeader\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\"`\n\x18\x43reateDownloadUrlRequest\x12\x0c\n\x04path\x18\x01 \x01(\t:6\xe2?3\n1com.databricks.rpc.RPC[CreateDownloadUrlResponse]\"\x82\x01\n\x19\x43reateDownloadUrlResponse\x12\x0b\n\x03url\x18\x01 \x01(\t\x12.\n\x07headers\x18\x02 \x03(\x0b\x32\x1d.mlflow.filesystem.HttpHeader:(\xe2?%\n#com.databricks.rpc.DoNotLogContents\"\\\n\x16\x43reateUploadUrlRequest\x12\x0c\n\x04path\x18\x01 \x01(\t:4\xe2?1\n/com.databricks.rpc.RPC[CreateUploadUrlResponse]\"\x80\x01\n\x17\x43reateUploadUrlResponse\x12\x0b\n\x03url\x18\x01 \x01(\t\x12.\n\x07headers\x18\x02 \x03(\x0b\x32\x1d.mlflow.filesystem.HttpHeader:(\xe2?%\n#com.databricks.rpc.DoNotLogContents\"\x96\x01\n\x0e\x44irectoryEntry\x12\x0c\n\x04path\x18\x01 \x01(\t\x12\x14\n\x0cis_directory\x18\x02 \x01(\x08\x12\x11\n\tfile_size\x18\x03 \x01(\x03\x12\x15\n\rlast_modified\x18\x04 \x01(\x03\x12\x0c\n\x04name\x18\x05 \x01(\t:(\xe2?%\n#com.databricks.rpc.DoNotLogContents\"\x8f\x01\n\x15ListDirectoryResponse\x12\x33\n\x08\x63ontents\x18\x01 \x03(\x0b\x32!.mlflow.filesystem.DirectoryEntry\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t:(\xe2?%\n#com.databricks.rpc.DoNotLogContents2\xcb\x02\n\x11\x46ilesystemService\x12\x9d\x01\n\x11\x43reateDownloadUrl\x12+.mlflow.filesystem.CreateDownloadUrlRequest\x1a,.mlflow.filesystem.CreateDownloadUrlResponse\"-\xf2\x86\x19)\n%\n\x04POST\x12\x17/fs/create-download-url\x1a\x04\x08\x02\x10\x00\x10\x03\x12\x95\x01\n\x0f\x43reateUploadUrl\x12).mlflow.filesystem.CreateUploadUrlRequest\x1a*.mlflow.filesystem.CreateUploadUrlResponse\"+\xf2\x86\x19\'\n#\n\x04POST\x12\x15/fs/create-upload-url\x1a\x04\x08\x02\x10\x00\x10\x03\x42\x30\n#com.databricks.api.proto.filesystem\x90\x01\x01\xa0\x01\x01\xe2?\x02\x10\x01')

  _globals = globals()
  _builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
  _builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'databricks_filesystem_service_pb2', _globals)
  if not _descriptor._USE_C_DESCRIPTORS:
    _globals['DESCRIPTOR']._loaded_options = None
    _globals['DESCRIPTOR']._serialized_options = b'\n#com.databricks.api.proto.filesystem\220\001\001\240\001\001\342?\002\020\001'
    _globals['_CREATEDOWNLOADURLREQUEST']._loaded_options = None
    _globals['_CREATEDOWNLOADURLREQUEST']._serialized_options = b'\342?3\n1com.databricks.rpc.RPC[CreateDownloadUrlResponse]'
    _globals['_CREATEDOWNLOADURLRESPONSE']._loaded_options = None
    _globals['_CREATEDOWNLOADURLRESPONSE']._serialized_options = b'\342?%\n#com.databricks.rpc.DoNotLogContents'
    _globals['_CREATEUPLOADURLREQUEST']._loaded_options = None
    _globals['_CREATEUPLOADURLREQUEST']._serialized_options = b'\342?1\n/com.databricks.rpc.RPC[CreateUploadUrlResponse]'
    _globals['_CREATEUPLOADURLRESPONSE']._loaded_options = None
    _globals['_CREATEUPLOADURLRESPONSE']._serialized_options = b'\342?%\n#com.databricks.rpc.DoNotLogContents'
    _globals['_DIRECTORYENTRY']._loaded_options = None
    _globals['_DIRECTORYENTRY']._serialized_options = b'\342?%\n#com.databricks.rpc.DoNotLogContents'
    _globals['_LISTDIRECTORYRESPONSE']._loaded_options = None
    _globals['_LISTDIRECTORYRESPONSE']._serialized_options = b'\342?%\n#com.databricks.rpc.DoNotLogContents'
    _globals['_FILESYSTEMSERVICE'].methods_by_name['CreateDownloadUrl']._loaded_options = None
    _globals['_FILESYSTEMSERVICE'].methods_by_name['CreateDownloadUrl']._serialized_options = b'\362\206\031)\n%\n\004POST\022\027/fs/create-download-url\032\004\010\002\020\000\020\003'
    _globals['_FILESYSTEMSERVICE'].methods_by_name['CreateUploadUrl']._loaded_options = None
    _globals['_FILESYSTEMSERVICE'].methods_by_name['CreateUploadUrl']._serialized_options = b'\362\206\031\'\n#\n\004POST\022\025/fs/create-upload-url\032\004\010\002\020\000\020\003'
    _globals['_HTTPHEADER']._serialized_start=99
    _globals['_HTTPHEADER']._serialized_end=140
    _globals['_CREATEDOWNLOADURLREQUEST']._serialized_start=142
    _globals['_CREATEDOWNLOADURLREQUEST']._serialized_end=238
    _globals['_CREATEDOWNLOADURLRESPONSE']._serialized_start=241
    _globals['_CREATEDOWNLOADURLRESPONSE']._serialized_end=371
    _globals['_CREATEUPLOADURLREQUEST']._serialized_start=373
    _globals['_CREATEUPLOADURLREQUEST']._serialized_end=465
    _globals['_CREATEUPLOADURLRESPONSE']._serialized_start=468
    _globals['_CREATEUPLOADURLRESPONSE']._serialized_end=596
    _globals['_DIRECTORYENTRY']._serialized_start=599
    _globals['_DIRECTORYENTRY']._serialized_end=749
    _globals['_LISTDIRECTORYRESPONSE']._serialized_start=752
    _globals['_LISTDIRECTORYRESPONSE']._serialized_end=895
    _globals['_FILESYSTEMSERVICE']._serialized_start=898
    _globals['_FILESYSTEMSERVICE']._serialized_end=1229
  _builder.BuildServices(DESCRIPTOR, 'databricks_filesystem_service_pb2', _globals)
  # @@protoc_insertion_point(module_scope)

else:
  # -*- coding: utf-8 -*-
  # Generated by the protocol buffer compiler.  DO NOT EDIT!
  # source: databricks_filesystem_service.proto
  """Generated protocol buffer code."""
  from google.protobuf import descriptor as _descriptor
  from google.protobuf import descriptor_pool as _descriptor_pool
  from google.protobuf import message as _message
  from google.protobuf import reflection as _reflection
  from google.protobuf import symbol_database as _symbol_database
  from google.protobuf import service as _service
  from google.protobuf import service_reflection
  # @@protoc_insertion_point(imports)

  _sym_db = _symbol_database.Default()


  from . import databricks_pb2 as databricks__pb2
  from .scalapb import scalapb_pb2 as scalapb_dot_scalapb__pb2


  DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n#databricks_filesystem_service.proto\x12\x11mlflow.filesystem\x1a\x10\x64\x61tabricks.proto\x1a\x15scalapb/scalapb.proto\")\n\nHttpHeader\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\"`\n\x18\x43reateDownloadUrlRequest\x12\x0c\n\x04path\x18\x01 \x01(\t:6\xe2?3\n1com.databricks.rpc.RPC[CreateDownloadUrlResponse]\"\x82\x01\n\x19\x43reateDownloadUrlResponse\x12\x0b\n\x03url\x18\x01 \x01(\t\x12.\n\x07headers\x18\x02 \x03(\x0b\x32\x1d.mlflow.filesystem.HttpHeader:(\xe2?%\n#com.databricks.rpc.DoNotLogContents\"\\\n\x16\x43reateUploadUrlRequest\x12\x0c\n\x04path\x18\x01 \x01(\t:4\xe2?1\n/com.databricks.rpc.RPC[CreateUploadUrlResponse]\"\x80\x01\n\x17\x43reateUploadUrlResponse\x12\x0b\n\x03url\x18\x01 \x01(\t\x12.\n\x07headers\x18\x02 \x03(\x0b\x32\x1d.mlflow.filesystem.HttpHeader:(\xe2?%\n#com.databricks.rpc.DoNotLogContents\"\x96\x01\n\x0e\x44irectoryEntry\x12\x0c\n\x04path\x18\x01 \x01(\t\x12\x14\n\x0cis_directory\x18\x02 \x01(\x08\x12\x11\n\tfile_size\x18\x03 \x01(\x03\x12\x15\n\rlast_modified\x18\x04 \x01(\x03\x12\x0c\n\x04name\x18\x05 \x01(\t:(\xe2?%\n#com.databricks.rpc.DoNotLogContents\"\x8f\x01\n\x15ListDirectoryResponse\x12\x33\n\x08\x63ontents\x18\x01 \x03(\x0b\x32!.mlflow.filesystem.DirectoryEntry\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t:(\xe2?%\n#com.databricks.rpc.DoNotLogContents2\xcb\x02\n\x11\x46ilesystemService\x12\x9d\x01\n\x11\x43reateDownloadUrl\x12+.mlflow.filesystem.CreateDownloadUrlRequest\x1a,.mlflow.filesystem.CreateDownloadUrlResponse\"-\xf2\x86\x19)\n%\n\x04POST\x12\x17/fs/create-download-url\x1a\x04\x08\x02\x10\x00\x10\x03\x12\x95\x01\n\x0f\x43reateUploadUrl\x12).mlflow.filesystem.CreateUploadUrlRequest\x1a*.mlflow.filesystem.CreateUploadUrlResponse\"+\xf2\x86\x19\'\n#\n\x04POST\x12\x15/fs/create-upload-url\x1a\x04\x08\x02\x10\x00\x10\x03\x42\x30\n#com.databricks.api.proto.filesystem\x90\x01\x01\xa0\x01\x01\xe2?\x02\x10\x01')



  _HTTPHEADER = DESCRIPTOR.message_types_by_name['HttpHeader']
  _CREATEDOWNLOADURLREQUEST = DESCRIPTOR.message_types_by_name['CreateDownloadUrlRequest']
  _CREATEDOWNLOADURLRESPONSE = DESCRIPTOR.message_types_by_name['CreateDownloadUrlResponse']
  _CREATEUPLOADURLREQUEST = DESCRIPTOR.message_types_by_name['CreateUploadUrlRequest']
  _CREATEUPLOADURLRESPONSE = DESCRIPTOR.message_types_by_name['CreateUploadUrlResponse']
  _DIRECTORYENTRY = DESCRIPTOR.message_types_by_name['DirectoryEntry']
  _LISTDIRECTORYRESPONSE = DESCRIPTOR.message_types_by_name['ListDirectoryResponse']
  HttpHeader = _reflection.GeneratedProtocolMessageType('HttpHeader', (_message.Message,), {
    'DESCRIPTOR' : _HTTPHEADER,
    '__module__' : 'databricks_filesystem_service_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.filesystem.HttpHeader)
    })
  _sym_db.RegisterMessage(HttpHeader)

  CreateDownloadUrlRequest = _reflection.GeneratedProtocolMessageType('CreateDownloadUrlRequest', (_message.Message,), {
    'DESCRIPTOR' : _CREATEDOWNLOADURLREQUEST,
    '__module__' : 'databricks_filesystem_service_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.filesystem.CreateDownloadUrlRequest)
    })
  _sym_db.RegisterMessage(CreateDownloadUrlRequest)

  CreateDownloadUrlResponse = _reflection.GeneratedProtocolMessageType('CreateDownloadUrlResponse', (_message.Message,), {
    'DESCRIPTOR' : _CREATEDOWNLOADURLRESPONSE,
    '__module__' : 'databricks_filesystem_service_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.filesystem.CreateDownloadUrlResponse)
    })
  _sym_db.RegisterMessage(CreateDownloadUrlResponse)

  CreateUploadUrlRequest = _reflection.GeneratedProtocolMessageType('CreateUploadUrlRequest', (_message.Message,), {
    'DESCRIPTOR' : _CREATEUPLOADURLREQUEST,
    '__module__' : 'databricks_filesystem_service_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.filesystem.CreateUploadUrlRequest)
    })
  _sym_db.RegisterMessage(CreateUploadUrlRequest)

  CreateUploadUrlResponse = _reflection.GeneratedProtocolMessageType('CreateUploadUrlResponse', (_message.Message,), {
    'DESCRIPTOR' : _CREATEUPLOADURLRESPONSE,
    '__module__' : 'databricks_filesystem_service_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.filesystem.CreateUploadUrlResponse)
    })
  _sym_db.RegisterMessage(CreateUploadUrlResponse)

  DirectoryEntry = _reflection.GeneratedProtocolMessageType('DirectoryEntry', (_message.Message,), {
    'DESCRIPTOR' : _DIRECTORYENTRY,
    '__module__' : 'databricks_filesystem_service_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.filesystem.DirectoryEntry)
    })
  _sym_db.RegisterMessage(DirectoryEntry)

  ListDirectoryResponse = _reflection.GeneratedProtocolMessageType('ListDirectoryResponse', (_message.Message,), {
    'DESCRIPTOR' : _LISTDIRECTORYRESPONSE,
    '__module__' : 'databricks_filesystem_service_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.filesystem.ListDirectoryResponse)
    })
  _sym_db.RegisterMessage(ListDirectoryResponse)

  _FILESYSTEMSERVICE = DESCRIPTOR.services_by_name['FilesystemService']
  if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    DESCRIPTOR._serialized_options = b'\n#com.databricks.api.proto.filesystem\220\001\001\240\001\001\342?\002\020\001'
    _CREATEDOWNLOADURLREQUEST._options = None
    _CREATEDOWNLOADURLREQUEST._serialized_options = b'\342?3\n1com.databricks.rpc.RPC[CreateDownloadUrlResponse]'
    _CREATEDOWNLOADURLRESPONSE._options = None
    _CREATEDOWNLOADURLRESPONSE._serialized_options = b'\342?%\n#com.databricks.rpc.DoNotLogContents'
    _CREATEUPLOADURLREQUEST._options = None
    _CREATEUPLOADURLREQUEST._serialized_options = b'\342?1\n/com.databricks.rpc.RPC[CreateUploadUrlResponse]'
    _CREATEUPLOADURLRESPONSE._options = None
    _CREATEUPLOADURLRESPONSE._serialized_options = b'\342?%\n#com.databricks.rpc.DoNotLogContents'
    _DIRECTORYENTRY._options = None
    _DIRECTORYENTRY._serialized_options = b'\342?%\n#com.databricks.rpc.DoNotLogContents'
    _LISTDIRECTORYRESPONSE._options = None
    _LISTDIRECTORYRESPONSE._serialized_options = b'\342?%\n#com.databricks.rpc.DoNotLogContents'
    _FILESYSTEMSERVICE.methods_by_name['CreateDownloadUrl']._options = None
    _FILESYSTEMSERVICE.methods_by_name['CreateDownloadUrl']._serialized_options = b'\362\206\031)\n%\n\004POST\022\027/fs/create-download-url\032\004\010\002\020\000\020\003'
    _FILESYSTEMSERVICE.methods_by_name['CreateUploadUrl']._options = None
    _FILESYSTEMSERVICE.methods_by_name['CreateUploadUrl']._serialized_options = b'\362\206\031\'\n#\n\004POST\022\025/fs/create-upload-url\032\004\010\002\020\000\020\003'
    _HTTPHEADER._serialized_start=99
    _HTTPHEADER._serialized_end=140
    _CREATEDOWNLOADURLREQUEST._serialized_start=142
    _CREATEDOWNLOADURLREQUEST._serialized_end=238
    _CREATEDOWNLOADURLRESPONSE._serialized_start=241
    _CREATEDOWNLOADURLRESPONSE._serialized_end=371
    _CREATEUPLOADURLREQUEST._serialized_start=373
    _CREATEUPLOADURLREQUEST._serialized_end=465
    _CREATEUPLOADURLRESPONSE._serialized_start=468
    _CREATEUPLOADURLRESPONSE._serialized_end=596
    _DIRECTORYENTRY._serialized_start=599
    _DIRECTORYENTRY._serialized_end=749
    _LISTDIRECTORYRESPONSE._serialized_start=752
    _LISTDIRECTORYRESPONSE._serialized_end=895
    _FILESYSTEMSERVICE._serialized_start=898
    _FILESYSTEMSERVICE._serialized_end=1229
  FilesystemService = service_reflection.GeneratedServiceType('FilesystemService', (_service.Service,), dict(
    DESCRIPTOR = _FILESYSTEMSERVICE,
    __module__ = 'databricks_filesystem_service_pb2'
    ))

  FilesystemService_Stub = service_reflection.GeneratedServiceStubType('FilesystemService_Stub', (FilesystemService,), dict(
    DESCRIPTOR = _FILESYSTEMSERVICE,
    __module__ = 'databricks_filesystem_service_pb2'
    ))


  # @@protoc_insertion_point(module_scope)




import inspect

from packaging.version import Version


def convert_to_serializable(response):
    """
    Convert the response to a JSON serializable format.

    LangChain response objects often contains Pydantic objects, which causes an serialization
    error when the model is served behind REST endpoint.
    """
    import langchain

    # LangChain >= 0.3.0 uses Pydantic 2.x while < 0.3.0 is based on Pydantic 1.x.
    if Version(langchain.__version__) >= Version("0.3.0"):
        from pydantic import BaseModel

        if isinstance(response, BaseModel):
            return response.model_dump()
    else:
        from langchain_core.pydantic_v1 import BaseModel as LangChainBaseModel

        if isinstance(response, LangChainBaseModel):
            return response.dict()

    if inspect.isgenerator(response):
        return (convert_to_serializable(chunk) for chunk in response)
    elif isinstance(response, dict):
        return {k: convert_to_serializable(v) for k, v in response.items()}
    elif isinstance(response, list):
        return [convert_to_serializable(v) for v in response]
    elif isinstance(response, tuple):
        return tuple(convert_to_serializable(v) for v in response)

    return response



from mlflow.gateway.base_models import RequestModel, ResponseModel
from mlflow.types.chat import BaseRequestPayload
from mlflow.utils import IS_PYDANTIC_V2_OR_NEWER

_REQUEST_PAYLOAD_EXTRA_SCHEMA = {
    "example": {
        "prompt": "hello",
        "temperature": 0.0,
        "max_tokens": 64,
        "stop": ["END"],
        "n": 1,
    }
}


class RequestPayload(BaseRequestPayload, RequestModel):
    prompt: str
    model: str | None = None

    class Config:
        if IS_PYDANTIC_V2_OR_NEWER:
            json_schema_extra = _REQUEST_PAYLOAD_EXTRA_SCHEMA
        else:
            schema_extra = _REQUEST_PAYLOAD_EXTRA_SCHEMA


class Choice(ResponseModel):
    index: int
    text: str
    finish_reason: str | None = None


class CompletionsUsage(ResponseModel):
    prompt_tokens: int | None = None
    completion_tokens: int | None = None
    total_tokens: int | None = None


_RESPONSE_PAYLOAD_EXTRA_SCHEMA = {
    "example": {
        "id": "cmpl-123",
        "object": "text_completion",
        "created": 1589478378,
        "model": "gpt-4",
        "choices": [
            {"text": "Hello! I am an AI Assistant!", "index": 0, "finish_reason": "length"}
        ],
        "usage": {"prompt_tokens": 5, "completion_tokens": 7, "total_tokens": 12},
    }
}


class ResponsePayload(ResponseModel):
    id: str | None = None
    object: str = "text_completion"
    created: int
    model: str
    choices: list[Choice]
    usage: CompletionsUsage

    class Config:
        if IS_PYDANTIC_V2_OR_NEWER:
            json_schema_extra = _RESPONSE_PAYLOAD_EXTRA_SCHEMA
        else:
            schema_extra = _RESPONSE_PAYLOAD_EXTRA_SCHEMA


class StreamDelta(ResponseModel):
    role: str | None = None
    content: str | None = None


class StreamChoice(ResponseModel):
    index: int
    finish_reason: str | None = None
    text: str | None = None


_STREAM_RESPONSE_PAYLOAD_EXTRA_SCHEMA = {
    "example": {
        "id": "cmpl-123",
        "object": "text_completion",
        "created": 1589478378,
        "model": "gpt-4",
        "choices": [
            {
                "index": 6,
                "finish_reason": "stop",
                "delta": {"role": "assistant", "content": "you?"},
            }
        ],
    }
}


class StreamResponsePayload(ResponseModel):
    id: str | None = None
    object: str = "text_completion_chunk"
    created: int
    model: str
    choices: list[StreamChoice]

    class Config:
        if IS_PYDANTIC_V2_OR_NEWER:
            json_schema_extra = _STREAM_RESPONSE_PAYLOAD_EXTRA_SCHEMA
        else:
            schema_extra = _STREAM_RESPONSE_PAYLOAD_EXTRA_SCHEMA



import functools
import warnings

import mlflow
from mlflow.models.evaluation import evaluate as model_evaluate
from mlflow.utils.uri import is_databricks_uri


@functools.wraps(model_evaluate)
def evaluate(*args, **kwargs):
    tracking_uri = mlflow.get_tracking_uri()
    if is_databricks_uri(tracking_uri):
        warnings.warn(
            "The `mlflow.evaluate` API has been deprecated as of MLflow 3.0.0. "
            "Please use these new alternatives:\n\n"
            " - For traditional ML or deep learning models: Use `mlflow.models.evaluate`, "
            "which maintains full compatibility with the original `mlflow.evaluate` API.\n\n"
            " - For LLMs or GenAI applications: Use the new `mlflow.genai.evaluate` API, "
            "which offers enhanced features specifically designed for evaluating "
            "LLMs and GenAI applications.\n",
            FutureWarning,
        )
    return model_evaluate(*args, **kwargs)



import posixpath
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Callable

from mlflow.entities.file_info import FileInfo
from mlflow.protos.databricks_artifacts_pb2 import (
    DatabricksMlflowArtifactsService,
    GetCredentialsForLoggedModelDownload,
    GetCredentialsForLoggedModelUpload,
    GetCredentialsForRead,
    GetCredentialsForTraceDataDownload,
    GetCredentialsForTraceDataUpload,
    GetCredentialsForWrite,
)
from mlflow.protos.service_pb2 import (
    GetLoggedModel,
    GetRun,
    ListArtifacts,
    ListLoggedModelArtifacts,
    MlflowService,
)
from mlflow.utils.proto_json_utils import message_to_json
from mlflow.utils.uri import extract_and_normalize_path


class _CredentialType(Enum):
    READ = 1
    WRITE = 2


@dataclass
class HttpHeader:
    name: str
    value: str


@dataclass
class ArtifactCredentialInfo:
    signed_uri: str
    type: Any
    headers: list[HttpHeader] = field(default_factory=list)


@dataclass
class ListArtifactsPage:
    # List of files in the current page
    files: list[FileInfo]
    # Token to fetch the next page of files
    next_page_token: str | None = None

    @classmethod
    def empty(cls):
        return cls(files=[], next_page_token=None)


class _Resource(ABC):
    """
    Represents a resource that `DatabricksArtifactRepository` interacts with.
    """

    def __init__(self, id_: str, artifact_uri: str, call_endpoint: Callable[..., Any]):
        self.id = id_
        self.artifact_uri = artifact_uri
        self._call_endpoint = call_endpoint
        self._artifact_root = None
        self._relative_path = None

    @property
    def call_endpoint(self) -> Callable[..., Any]:
        return self._call_endpoint

    @property
    def artifact_root(self) -> str:
        if self._artifact_root is None:
            self._artifact_root = self.get_artifact_root()
        return self._artifact_root

    @property
    def relative_path(self) -> str:
        if self._relative_path is None:
            # Fetch the artifact root for the MLflow resource associated with `artifact_uri` and
            # compute the path of `artifact_uri` relative to the MLflow resource's artifact root
            # All operations performed on this artifact repository will be performed relative to
            # this computed location.
            artifact_repo_root_path = extract_and_normalize_path(self.artifact_uri)
            artifact_root_path = extract_and_normalize_path(self.artifact_root)
            # If the paths are equal, then use empty string over "./" for ListArtifact compatibility
            self._relative_path = (
                ""
                if artifact_root_path == artifact_repo_root_path
                else posixpath.relpath(artifact_repo_root_path, artifact_root_path)
            )
        return self._relative_path

    @abstractmethod
    def get_credentials(
        self,
        cred_type: _CredentialType,
        paths: list[str] | None = None,
        page_token: str | None = None,
    ) -> tuple[list[ArtifactCredentialInfo], str | None]:
        """
        Fetches read/write credentials for the specified paths.
        """

    @abstractmethod
    def get_artifact_root(self) -> str:
        """
        Get the artifact root URI of this resource.
        """

    @abstractmethod
    def _list_artifacts(
        self,
        path: str | None = None,
        page_token: str | None = None,
    ) -> ListArtifactsPage:
        """
        List artifacts under the specified path.
        """

    def list_artifacts(self, path: str | None = None) -> list[FileInfo]:
        """
        Handle pagination and return all artifacts under the specified path.
        """
        files: list[FileInfo] = []
        page_token: str | None = None
        while True:
            page = self._list_artifacts(path, page_token)
            files.extend(page.files)
            if len(page.files) == 0 or not page.next_page_token:
                break
            page_token = page.next_page_token

        return files


class _LoggedModel(_Resource):
    def get_credentials(
        self,
        cred_type: _CredentialType,
        paths: list[str] | None = None,
        page_token: str | None = None,
    ) -> tuple[list[ArtifactCredentialInfo], str | None]:
        api = (
            GetCredentialsForLoggedModelDownload
            if cred_type == _CredentialType.READ
            else GetCredentialsForLoggedModelUpload
        )
        payload = api(paths=paths, page_token=page_token)
        response = self.call_endpoint(
            DatabricksMlflowArtifactsService,
            api,
            message_to_json(payload),
            path_params={"model_id": self.id},
        )
        credential_infos = [
            ArtifactCredentialInfo(
                signed_uri=c.credential_info.signed_uri,
                type=c.credential_info.type,
                headers=[HttpHeader(name=h.name, value=h.value) for h in c.credential_info.headers],
            )
            for c in response.credentials
        ]
        return credential_infos, response.next_page_token

    def get_artifact_root(self) -> str:
        json_body = message_to_json(GetLoggedModel(model_id=self.id))
        response = self.call_endpoint(
            MlflowService, GetLoggedModel, json_body, path_params={"model_id": self.id}
        )
        return response.model.info.artifact_uri

    def _list_artifacts(
        self,
        path: str | None = None,
        page_token: str | None = None,
    ) -> ListArtifactsPage:
        path = posixpath.join(self.relative_path, path) if path else self.relative_path
        json_body = message_to_json(
            ListLoggedModelArtifacts(page_token=page_token, artifact_directory_path=path)
        )
        response = self.call_endpoint(
            MlflowService, ListLoggedModelArtifacts, json_body, path_params={"model_id": self.id}
        )
        files = response.files
        # If `path` is a file, ListLoggedModelArtifacts returns a single list element with the
        # same name as `path`. The list_artifacts API expects us to return an empty list in this
        # case, so we do so here.
        if len(files) == 1 and files[0].path == path and not files[0].is_dir:
            return ListArtifactsPage.empty()

        return ListArtifactsPage(
            files=[
                FileInfo(
                    posixpath.relpath(f.path, self.relative_path),
                    f.is_dir,
                    None if f.is_dir else f.file_size,
                )
                for f in files
            ],
            next_page_token=response.next_page_token,
        )


class _Run(_Resource):
    def get_credentials(
        self,
        cred_type: _CredentialType,
        paths: list[str] | None = None,
        page_token: str | None = None,
    ) -> tuple[list[ArtifactCredentialInfo], str | None]:
        api = GetCredentialsForRead if cred_type == _CredentialType.READ else GetCredentialsForWrite
        json_body = api(run_id=self.id, path=paths, page_token=page_token)
        response = self.call_endpoint(
            DatabricksMlflowArtifactsService, api, message_to_json(json_body)
        )
        credential_infos = [
            ArtifactCredentialInfo(
                signed_uri=c.signed_uri,
                type=c.type,
                headers=[HttpHeader(name=h.name, value=h.value) for h in c.headers],
            )
            for c in response.credential_infos
        ]
        return credential_infos, response.next_page_token

    def get_artifact_root(self) -> str:
        json_body = message_to_json(GetRun(run_id=self.id))
        run_response = self.call_endpoint(MlflowService, GetRun, json_body)
        return run_response.run.info.artifact_uri

    def _list_artifacts(
        self,
        path: str | None = None,
        page_token: str | None = None,
    ) -> ListArtifactsPage:
        path = posixpath.join(self.relative_path, path) if path else self.relative_path
        json_body = message_to_json(
            ListArtifacts(run_id=self.id, path=path, page_token=page_token),
        )
        response = self.call_endpoint(MlflowService, ListArtifacts, json_body)
        files = response.files
        # If `path` is a file, ListArtifacts returns a single list element with the
        # same name as `path`. The list_artifacts API expects us to return an empty list in this
        # case, so we do so here.
        if len(files) == 1 and files[0].path == path and not files[0].is_dir:
            return ListArtifactsPage.empty()

        return ListArtifactsPage(
            files=[
                FileInfo(
                    posixpath.relpath(f.path, self.relative_path),
                    f.is_dir,
                    None if f.is_dir else f.file_size,
                )
                for f in files
            ],
            next_page_token=response.next_page_token,
        )


class _Trace(_Resource):
    def get_artifact_root(self) -> str:
        return None

    def get_credentials(
        self,
        cred_type: _CredentialType,
        paths: list[str] | None = None,
        page_token: str | None = None,
        timeout: int | None = None,
    ) -> tuple[list[ArtifactCredentialInfo], str | None]:
        res = self.call_endpoint(
            DatabricksMlflowArtifactsService,
            (
                GetCredentialsForTraceDataDownload
                if cred_type == _CredentialType.READ
                else GetCredentialsForTraceDataUpload
            ),
            path_params={"request_id": self.id},
            retry_timeout_seconds=timeout,
        )
        cred_inf = ArtifactCredentialInfo(
            signed_uri=res.credential_info.signed_uri,
            type=res.credential_info.type,
            headers=[HttpHeader(name=h.name, value=h.value) for h in res.credential_info.headers],
        )
        return [cred_inf], None

    def get_artifact_root(self) -> str:
        raise NotImplementedError

    def _list_artifacts(
        self,
        path: str | None = None,
        page_token: str | None = None,
    ) -> ListArtifactsPage:
        raise NotImplementedError



from scalapb import scalapb_pb2 as _scalapb_pb2
import databricks_pb2 as _databricks_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import service as _service
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class ArtifactCredentialType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    AZURE_SAS_URI: _ClassVar[ArtifactCredentialType]
    AWS_PRESIGNED_URL: _ClassVar[ArtifactCredentialType]
    GCP_SIGNED_URL: _ClassVar[ArtifactCredentialType]
    AZURE_ADLS_GEN2_SAS_URI: _ClassVar[ArtifactCredentialType]
AZURE_SAS_URI: ArtifactCredentialType
AWS_PRESIGNED_URL: ArtifactCredentialType
GCP_SIGNED_URL: ArtifactCredentialType
AZURE_ADLS_GEN2_SAS_URI: ArtifactCredentialType

class ArtifactCredentialInfo(_message.Message):
    __slots__ = ("run_id", "path", "signed_uri", "headers", "type")
    class HttpHeader(_message.Message):
        __slots__ = ("name", "value")
        NAME_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        name: str
        value: str
        def __init__(self, name: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    RUN_ID_FIELD_NUMBER: _ClassVar[int]
    PATH_FIELD_NUMBER: _ClassVar[int]
    SIGNED_URI_FIELD_NUMBER: _ClassVar[int]
    HEADERS_FIELD_NUMBER: _ClassVar[int]
    TYPE_FIELD_NUMBER: _ClassVar[int]
    run_id: str
    path: str
    signed_uri: str
    headers: _containers.RepeatedCompositeFieldContainer[ArtifactCredentialInfo.HttpHeader]
    type: ArtifactCredentialType
    def __init__(self, run_id: _Optional[str] = ..., path: _Optional[str] = ..., signed_uri: _Optional[str] = ..., headers: _Optional[_Iterable[_Union[ArtifactCredentialInfo.HttpHeader, _Mapping]]] = ..., type: _Optional[_Union[ArtifactCredentialType, str]] = ...) -> None: ...

class LoggedModelArtifactCredential(_message.Message):
    __slots__ = ("model_id", "credential_info")
    MODEL_ID_FIELD_NUMBER: _ClassVar[int]
    CREDENTIAL_INFO_FIELD_NUMBER: _ClassVar[int]
    model_id: str
    credential_info: ArtifactCredentialInfo
    def __init__(self, model_id: _Optional[str] = ..., credential_info: _Optional[_Union[ArtifactCredentialInfo, _Mapping]] = ...) -> None: ...

class GetCredentialsForRead(_message.Message):
    __slots__ = ("run_id", "path", "page_token")
    class Response(_message.Message):
        __slots__ = ("credential_infos", "next_page_token")
        CREDENTIAL_INFOS_FIELD_NUMBER: _ClassVar[int]
        NEXT_PAGE_TOKEN_FIELD_NUMBER: _ClassVar[int]
        credential_infos: _containers.RepeatedCompositeFieldContainer[ArtifactCredentialInfo]
        next_page_token: str
        def __init__(self, credential_infos: _Optional[_Iterable[_Union[ArtifactCredentialInfo, _Mapping]]] = ..., next_page_token: _Optional[str] = ...) -> None: ...
    RUN_ID_FIELD_NUMBER: _ClassVar[int]
    PATH_FIELD_NUMBER: _ClassVar[int]
    PAGE_TOKEN_FIELD_NUMBER: _ClassVar[int]
    run_id: str
    path: _containers.RepeatedScalarFieldContainer[str]
    page_token: str
    def __init__(self, run_id: _Optional[str] = ..., path: _Optional[_Iterable[str]] = ..., page_token: _Optional[str] = ...) -> None: ...

class GetCredentialsForWrite(_message.Message):
    __slots__ = ("run_id", "path", "page_token")
    class Response(_message.Message):
        __slots__ = ("credential_infos", "next_page_token")
        CREDENTIAL_INFOS_FIELD_NUMBER: _ClassVar[int]
        NEXT_PAGE_TOKEN_FIELD_NUMBER: _ClassVar[int]
        credential_infos: _containers.RepeatedCompositeFieldContainer[ArtifactCredentialInfo]
        next_page_token: str
        def __init__(self, credential_infos: _Optional[_Iterable[_Union[ArtifactCredentialInfo, _Mapping]]] = ..., next_page_token: _Optional[str] = ...) -> None: ...
    RUN_ID_FIELD_NUMBER: _ClassVar[int]
    PATH_FIELD_NUMBER: _ClassVar[int]
    PAGE_TOKEN_FIELD_NUMBER: _ClassVar[int]
    run_id: str
    path: _containers.RepeatedScalarFieldContainer[str]
    page_token: str
    def __init__(self, run_id: _Optional[str] = ..., path: _Optional[_Iterable[str]] = ..., page_token: _Optional[str] = ...) -> None: ...

class CreateMultipartUpload(_message.Message):
    __slots__ = ("run_id", "path", "num_parts")
    class Response(_message.Message):
        __slots__ = ("upload_id", "upload_credential_infos", "abort_credential_info")
        UPLOAD_ID_FIELD_NUMBER: _ClassVar[int]
        UPLOAD_CREDENTIAL_INFOS_FIELD_NUMBER: _ClassVar[int]
        ABORT_CREDENTIAL_INFO_FIELD_NUMBER: _ClassVar[int]
        upload_id: str
        upload_credential_infos: _containers.RepeatedCompositeFieldContainer[ArtifactCredentialInfo]
        abort_credential_info: ArtifactCredentialInfo
        def __init__(self, upload_id: _Optional[str] = ..., upload_credential_infos: _Optional[_Iterable[_Union[ArtifactCredentialInfo, _Mapping]]] = ..., abort_credential_info: _Optional[_Union[ArtifactCredentialInfo, _Mapping]] = ...) -> None: ...
    RUN_ID_FIELD_NUMBER: _ClassVar[int]
    PATH_FIELD_NUMBER: _ClassVar[int]
    NUM_PARTS_FIELD_NUMBER: _ClassVar[int]
    run_id: str
    path: str
    num_parts: int
    def __init__(self, run_id: _Optional[str] = ..., path: _Optional[str] = ..., num_parts: _Optional[int] = ...) -> None: ...

class PartEtag(_message.Message):
    __slots__ = ("part_number", "etag")
    PART_NUMBER_FIELD_NUMBER: _ClassVar[int]
    ETAG_FIELD_NUMBER: _ClassVar[int]
    part_number: int
    etag: str
    def __init__(self, part_number: _Optional[int] = ..., etag: _Optional[str] = ...) -> None: ...

class CompleteMultipartUpload(_message.Message):
    __slots__ = ("run_id", "path", "upload_id", "part_etags")
    class Response(_message.Message):
        __slots__ = ()
        def __init__(self) -> None: ...
    RUN_ID_FIELD_NUMBER: _ClassVar[int]
    PATH_FIELD_NUMBER: _ClassVar[int]
    UPLOAD_ID_FIELD_NUMBER: _ClassVar[int]
    PART_ETAGS_FIELD_NUMBER: _ClassVar[int]
    run_id: str
    path: str
    upload_id: str
    part_etags: _containers.RepeatedCompositeFieldContainer[PartEtag]
    def __init__(self, run_id: _Optional[str] = ..., path: _Optional[str] = ..., upload_id: _Optional[str] = ..., part_etags: _Optional[_Iterable[_Union[PartEtag, _Mapping]]] = ...) -> None: ...

class GetPresignedUploadPartUrl(_message.Message):
    __slots__ = ("run_id", "path", "upload_id", "part_number")
    class Response(_message.Message):
        __slots__ = ("upload_credential_info",)
        UPLOAD_CREDENTIAL_INFO_FIELD_NUMBER: _ClassVar[int]
        upload_credential_info: ArtifactCredentialInfo
        def __init__(self, upload_credential_info: _Optional[_Union[ArtifactCredentialInfo, _Mapping]] = ...) -> None: ...
    RUN_ID_FIELD_NUMBER: _ClassVar[int]
    PATH_FIELD_NUMBER: _ClassVar[int]
    UPLOAD_ID_FIELD_NUMBER: _ClassVar[int]
    PART_NUMBER_FIELD_NUMBER: _ClassVar[int]
    run_id: str
    path: str
    upload_id: str
    part_number: int
    def __init__(self, run_id: _Optional[str] = ..., path: _Optional[str] = ..., upload_id: _Optional[str] = ..., part_number: _Optional[int] = ...) -> None: ...

class GetCredentialsForTraceDataDownload(_message.Message):
    __slots__ = ("request_id",)
    class Response(_message.Message):
        __slots__ = ("credential_info",)
        CREDENTIAL_INFO_FIELD_NUMBER: _ClassVar[int]
        credential_info: ArtifactCredentialInfo
        def __init__(self, credential_info: _Optional[_Union[ArtifactCredentialInfo, _Mapping]] = ...) -> None: ...
    REQUEST_ID_FIELD_NUMBER: _ClassVar[int]
    request_id: str
    def __init__(self, request_id: _Optional[str] = ...) -> None: ...

class GetCredentialsForTraceDataUpload(_message.Message):
    __slots__ = ("request_id",)
    class Response(_message.Message):
        __slots__ = ("credential_info",)
        CREDENTIAL_INFO_FIELD_NUMBER: _ClassVar[int]
        credential_info: ArtifactCredentialInfo
        def __init__(self, credential_info: _Optional[_Union[ArtifactCredentialInfo, _Mapping]] = ...) -> None: ...
    REQUEST_ID_FIELD_NUMBER: _ClassVar[int]
    request_id: str
    def __init__(self, request_id: _Optional[str] = ...) -> None: ...

class GetCredentialsForLoggedModelUpload(_message.Message):
    __slots__ = ("model_id", "paths", "page_token")
    class Response(_message.Message):
        __slots__ = ("credentials", "next_page_token")
        CREDENTIALS_FIELD_NUMBER: _ClassVar[int]
        NEXT_PAGE_TOKEN_FIELD_NUMBER: _ClassVar[int]
        credentials: _containers.RepeatedCompositeFieldContainer[LoggedModelArtifactCredential]
        next_page_token: str
        def __init__(self, credentials: _Optional[_Iterable[_Union[LoggedModelArtifactCredential, _Mapping]]] = ..., next_page_token: _Optional[str] = ...) -> None: ...
    MODEL_ID_FIELD_NUMBER: _ClassVar[int]
    PATHS_FIELD_NUMBER: _ClassVar[int]
    PAGE_TOKEN_FIELD_NUMBER: _ClassVar[int]
    model_id: str
    paths: _containers.RepeatedScalarFieldContainer[str]
    page_token: str
    def __init__(self, model_id: _Optional[str] = ..., paths: _Optional[_Iterable[str]] = ..., page_token: _Optional[str] = ...) -> None: ...

class GetCredentialsForLoggedModelDownload(_message.Message):
    __slots__ = ("model_id", "paths", "page_token")
    class Response(_message.Message):
        __slots__ = ("credentials", "next_page_token")
        CREDENTIALS_FIELD_NUMBER: _ClassVar[int]
        NEXT_PAGE_TOKEN_FIELD_NUMBER: _ClassVar[int]
        credentials: _containers.RepeatedCompositeFieldContainer[LoggedModelArtifactCredential]
        next_page_token: str
        def __init__(self, credentials: _Optional[_Iterable[_Union[LoggedModelArtifactCredential, _Mapping]]] = ..., next_page_token: _Optional[str] = ...) -> None: ...
    MODEL_ID_FIELD_NUMBER: _ClassVar[int]
    PATHS_FIELD_NUMBER: _ClassVar[int]
    PAGE_TOKEN_FIELD_NUMBER: _ClassVar[int]
    model_id: str
    paths: _containers.RepeatedScalarFieldContainer[str]
    page_token: str
    def __init__(self, model_id: _Optional[str] = ..., paths: _Optional[_Iterable[str]] = ..., page_token: _Optional[str] = ...) -> None: ...

class DatabricksMlflowArtifactsService(_service.service): ...

class DatabricksMlflowArtifactsService_Stub(DatabricksMlflowArtifactsService): ...



import json
import logging
from dataclasses import dataclass
from types import FunctionType
from typing import Any, Callable, ParamSpec, TypeVar

from mlflow.entities._job import Job as JobEntity
from mlflow.exceptions import MlflowException
from mlflow.server.handlers import _get_job_store

_logger = logging.getLogger(__name__)

P = ParamSpec("P")
R = TypeVar("R")


class TransientError(RuntimeError):
    """
    Raise `TransientError` in a job to trigger job retry
    """

    def __init__(self, origin_error: Exception):
        super().__init__()
        self._origin_error = origin_error

    @property
    def origin_error(self) -> Exception:
        return self._origin_error


@dataclass
class JobFunctionMetadata:
    fn_fullname: str
    max_workers: int
    use_process: bool


def job(
    max_workers: int,
    use_process: bool = True,
) -> Callable[[Callable[P, R]], Callable[P, R]]:
    """
    The decorator for the custom job function for setting max parallel workers that
    the job function can use.

    Args:
        max_workers: The maximum number of workers that are allowed to run the jobs
            using this job function.
        use_process: (optional) Specify whether to run the job in an individual process.
            If the job uses environment variables (e.g. API keys),
            it should be run in an individual process to isolate the environment variable settings.
            Default value is True.
    """

    def decorator(fn: Callable[P, R]) -> Callable[P, R]:
        fn._job_fn_metadata = JobFunctionMetadata(
            fn_fullname=f"{fn.__module__}.{fn.__name__}",
            max_workers=max_workers,
            use_process=use_process,
        )
        return fn

    return decorator


def submit_job(
    function: Callable[..., Any],
    params: dict[str, Any],
    timeout: float | None = None,
) -> JobEntity:
    """
    Submit a job to the job queue. The job is executed at most once.
    If the MLflow server crashes while the job is pending or running,
    it is rescheduled on restart.

    Note:
        This is a server-side API and requires the MLflow server to configure
        the backend store URI to a database URI.

    Args:
        function: The job function, it must be a python module-level function,
            and all params and return value must be JSON-serializable.
            The function can raise `TransientError` in order to trigger
            job retry, you can set `MLFLOW_SERVER_JOB_TRANSIENT_ERROR_MAX_RETRIES`
            to configure maximum allowed retries for transient errors
            and set `MLFLOW_SERVER_JOB_TRANSIENT_ERROR_RETRY_BASE_DELAY` to
            configure base retry delay in seconds.

            The function must be decorated by `mlflow.server.jobs.job_function` decorator.
        params: The params to be passed to the job function.
        timeout: (optional) The job execution timeout, default None (no timeout)

    Returns:
        The job entity. You can call `get_job` API by the job id to get
        the updated job entity.
    """
    from mlflow.environment_variables import MLFLOW_SERVER_ENABLE_JOB_EXECUTION
    from mlflow.server.jobs.utils import (
        _check_requirements,
        _get_or_init_huey_instance,
        _validate_function_parameters,
    )

    _check_requirements()

    if not MLFLOW_SERVER_ENABLE_JOB_EXECUTION.get():
        raise MlflowException(
            "Mlflow server job execution feature is not enabled, please set "
            "environment variable 'MLFLOW_SERVER_ENABLE_JOB_EXECUTION' to 'true' to enable it."
        )

    if not (isinstance(function, FunctionType) and "." not in function.__qualname__):
        raise MlflowException("The job function must be a python global function.")

    func_fullname = f"{function.__module__}.{function.__name__}"

    if not hasattr(function, "_job_fn_metadata"):
        raise MlflowException(
            f"The job function {func_fullname} is not decorated by "
            "'mlflow.server.jobs.job_function'."
        )

    # Validate that required parameters are provided
    _validate_function_parameters(function, params)

    job_store = _get_job_store()
    serialized_params = json.dumps(params)
    job = job_store.create_job(func_fullname, serialized_params, timeout)

    # enqueue job
    _get_or_init_huey_instance(func_fullname).submit_task(
        job.job_id,
        function,
        params,
        timeout,
    )

    return job


def get_job(job_id: str) -> JobEntity:
    """
    Get the job entity by the job id.

    Note:
        This is a server-side API, and it requires MLflow server configures
        backend store URI to database URI.

    Args:
        job_id: The job id.

    Returns:
        The job entity. If the job does not exist, error is raised.
    """
    job_store = _get_job_store()
    return job_store.get_job(job_id)



"""
Utilities for dealing with artifacts in the context of a Run.
"""

import os
import pathlib
import posixpath
import tempfile
import urllib.parse
import uuid
from typing import Any

from mlflow.exceptions import MlflowException
from mlflow.protos.databricks_pb2 import INVALID_PARAMETER_VALUE
from mlflow.store.artifact.artifact_repository_registry import get_artifact_repository
from mlflow.store.artifact.dbfs_artifact_repo import DbfsRestArtifactRepository
from mlflow.store.artifact.models_artifact_repo import ModelsArtifactRepository
from mlflow.tracking._tracking_service.utils import _get_store
from mlflow.utils.file_utils import path_to_local_file_uri
from mlflow.utils.os import is_windows
from mlflow.utils.uri import add_databricks_profile_info_to_artifact_uri, append_to_uri_path


def get_artifact_uri(run_id, artifact_path=None, tracking_uri=None):
    """Get the absolute URI of the specified artifact in the specified run. If `path` is not
    specified the artifact root URI of the specified run will be returned; calls to ``log_artifact``
    and ``log_artifacts`` write artifact(s) to subdirectories of the artifact root URI.

    Args:
        run_id: The ID of the run for which to obtain an absolute artifact URI.
        artifact_path: The run-relative artifact path. For example,
            ``path/to/artifact``. If unspecified, the artifact root URI for the
            specified run will be returned.
        tracking_uri: The tracking URI from which to get the run and its artifact location. If
            not given, the current default tracking URI is used.

    Returns:
        An *absolute* URI referring to the specified artifact or the specified run's artifact
        root. For example, if an artifact path is provided and the specified run uses an
        S3-backed  store, this may be a uri of the form
        ``s3://<bucket_name>/path/to/artifact/root/path/to/artifact``. If an artifact path
        is not provided and the specified run uses an S3-backed store, this may be a URI of
        the form ``s3://<bucket_name>/path/to/artifact/root``.

    """
    if not run_id:
        raise MlflowException(
            message="A run_id must be specified in order to obtain an artifact uri!",
            error_code=INVALID_PARAMETER_VALUE,
        )

    store = _get_store(tracking_uri)
    run = store.get_run(run_id)
    # Maybe move this method to RunsArtifactRepository so the circular dependency is clearer.
    assert urllib.parse.urlparse(run.info.artifact_uri).scheme != "runs"  # avoid an infinite loop
    if artifact_path is None:
        return run.info.artifact_uri
    else:
        return append_to_uri_path(run.info.artifact_uri, artifact_path)


# TODO: This would be much simpler if artifact_repo.download_artifacts could take the absolute path
# or no path.
def _get_root_uri_and_artifact_path(artifact_uri):
    """Parse the artifact_uri to get the root_uri and artifact_path.

    Args:
        artifact_uri: The *absolute* URI of the artifact.
    """
    if os.path.exists(artifact_uri):
        if not is_windows():
            # If we're dealing with local files, just reference the direct pathing.
            # non-nt-based file systems can directly reference path information, while nt-based
            # systems need to url-encode special characters in directory listings to be able to
            # resolve them (i.e., spaces converted to %20 within a file name or path listing)
            root_uri = os.path.dirname(artifact_uri)
            artifact_path = os.path.basename(artifact_uri)
            return root_uri, artifact_path
        else:  # if we're dealing with nt-based systems, we need to utilize pathname2url to encode.
            artifact_uri = path_to_local_file_uri(artifact_uri)

    parsed_uri = urllib.parse.urlparse(str(artifact_uri))
    prefix = ""
    if parsed_uri.scheme and not parsed_uri.path.startswith("/"):
        # relative path is a special case, urllib does not reconstruct it properly
        prefix = parsed_uri.scheme + ":"
        parsed_uri = parsed_uri._replace(scheme="")

    # For models:/ URIs, it doesn't make sense to initialize a ModelsArtifactRepository with only
    # the model name portion of the URI, then call download_artifacts with the version info.
    if ModelsArtifactRepository.is_models_uri(artifact_uri):
        root_uri, artifact_path = ModelsArtifactRepository.split_models_uri(artifact_uri)
    else:
        artifact_path = posixpath.basename(parsed_uri.path)
        parsed_uri = parsed_uri._replace(path=posixpath.dirname(parsed_uri.path))
        root_uri = prefix + urllib.parse.urlunparse(parsed_uri)
    return root_uri, artifact_path


def _download_artifact_from_uri(
    artifact_uri: str,
    output_path: str | None = None,
    lineage_header_info: dict[str, Any] | None = None,
    tracking_uri: str | None = None,
    registry_uri: str | None = None,
) -> str:
    """
    Args:
        artifact_uri: The *absolute* URI of the artifact to download.
        output_path: The local filesystem path to which to download the artifact. If unspecified,
            a local output path will be created.
        lineage_header_info: The model lineage header info to be consumed by lineage services.
        tracking_uri: The tracking URI to be used when downloading artifacts.
        registry_uri: The registry URI to be used when downloading artifacts.
    """
    root_uri, artifact_path = _get_root_uri_and_artifact_path(artifact_uri)
    repo = get_artifact_repository(
        artifact_uri=root_uri, tracking_uri=tracking_uri, registry_uri=registry_uri
    )

    try:
        if isinstance(repo, ModelsArtifactRepository):
            return repo.download_artifacts(
                artifact_path=artifact_path,
                dst_path=output_path,
                lineage_header_info=lineage_header_info,
            )
        return repo.download_artifacts(artifact_path=artifact_path, dst_path=output_path)
    except Exception as e:
        if artifact_uri.startswith("m-"):
            # When a Model ID like string is passed, suggest using 'models:/{artifact_uri}' instead.
            raise MlflowException(
                f"Invalid uri `{artifact_uri}` is passed. Maybe you meant 'models:/{artifact_uri}'?"
            ) from e
        raise


def _upload_artifact_to_uri(local_path, artifact_uri):
    """Uploads a local artifact (file) to a specified URI.

    Args:
        local_path: The local path of the file to upload.
        artifact_uri: The *absolute* URI of the path to upload the artifact to.

    """
    root_uri, artifact_path = _get_root_uri_and_artifact_path(artifact_uri)
    get_artifact_repository(artifact_uri=root_uri).log_artifact(local_path, artifact_path)


def _upload_artifacts_to_databricks(
    source, run_id, source_host_uri=None, target_databricks_profile_uri=None
):
    """Copy the artifacts from ``source`` to the destination Databricks workspace (DBFS) given by
    ``databricks_profile_uri`` or the current tracking URI.

    Args:
        source: Source location for the artifacts to copy.
        run_id: Run ID to associate the artifacts with.
        source_host_uri: Specifies the source artifact's host URI (e.g. Databricks tracking URI)
            if applicable. If not given, defaults to the current tracking URI.
        target_databricks_profile_uri: Specifies the destination Databricks host. If not given,
            defaults to the current tracking URI.

    Returns:
        The DBFS location in the target Databricks workspace the model files have been
        uploaded to.
    """

    with tempfile.TemporaryDirectory() as local_dir:
        source_with_profile = add_databricks_profile_info_to_artifact_uri(source, source_host_uri)
        _download_artifact_from_uri(source_with_profile, local_dir)
        dest_root = "dbfs:/databricks/mlflow/tmp-external-source/"
        dest_root_with_profile = add_databricks_profile_info_to_artifact_uri(
            dest_root, target_databricks_profile_uri
        )
        dest_repo = DbfsRestArtifactRepository(dest_root_with_profile)
        dest_artifact_path = run_id if run_id else uuid.uuid4().hex
        # Allow uploading from the same run id multiple times by randomizing a suffix
        if len(dest_repo.list_artifacts(dest_artifact_path)) > 0:
            dest_artifact_path = dest_artifact_path + "-" + uuid.uuid4().hex[0:4]
        dest_repo.log_artifacts(local_dir, artifact_path=dest_artifact_path)
        dirname = pathlib.PurePath(source).name  # innermost directory name
        return posixpath.join(dest_root, dest_artifact_path, dirname)  # new source



# MLflow webhook headers
WEBHOOK_SIGNATURE_HEADER = "X-MLflow-Signature"
WEBHOOK_TIMESTAMP_HEADER = "X-MLflow-Timestamp"
WEBHOOK_DELIVERY_ID_HEADER = "X-MLflow-Delivery-Id"

# Webhook signature version
WEBHOOK_SIGNATURE_VERSION = "v1"



"""add cascade deletion to trace tables foreign keys

Create Date: 2024-05-22 17:44:24.597019

"""

from alembic import op

from mlflow.store.tracking.dbmodels.models import SqlTraceInfo, SqlTraceMetadata, SqlTraceTag

# revision identifiers, used by Alembic.
revision = "5b0e9adcef9c"
down_revision = "867495a8f9d4"
branch_labels = None
depends_on = None


def upgrade():
    tables = [SqlTraceTag.__tablename__, SqlTraceMetadata.__tablename__]
    for table in tables:
        fk_tag_constraint_name = f"fk_{table}_request_id"
        # We have to use batch_alter_table as SQLite does not support
        # ALTER outside of a batch operation.
        with op.batch_alter_table(table, schema=None) as batch_op:
            batch_op.drop_constraint(fk_tag_constraint_name, type_="foreignkey")
            batch_op.create_foreign_key(
                fk_tag_constraint_name,
                SqlTraceInfo.__tablename__,
                ["request_id"],
                ["request_id"],
                # Add cascade deletion to the foreign key constraint.
                # This is the only change in this migration.
                ondelete="CASCADE",
            )


def downgrade():
    pass



from mlflow.protos.databricks_uc_registry_messages_pb2 import (
    ModelVersionDeploymentJobState as ProtoModelVersionDeploymentJobState,
)


class ModelVersionDeploymentJobRunState:
    """Enum for model version deployment state of an
    :py:class:`mlflow.entities.model_registry.ModelVersion`.
    """

    NO_VALID_DEPLOYMENT_JOB_FOUND = ProtoModelVersionDeploymentJobState.DeploymentJobRunState.Value(
        "NO_VALID_DEPLOYMENT_JOB_FOUND"
    )
    RUNNING = ProtoModelVersionDeploymentJobState.DeploymentJobRunState.Value("RUNNING")
    SUCCEEDED = ProtoModelVersionDeploymentJobState.DeploymentJobRunState.Value("SUCCEEDED")
    FAILED = ProtoModelVersionDeploymentJobState.DeploymentJobRunState.Value("FAILED")
    PENDING = ProtoModelVersionDeploymentJobState.DeploymentJobRunState.Value("PENDING")
    _STRING_TO_STATE = {
        k: ProtoModelVersionDeploymentJobState.DeploymentJobRunState.Value(k)
        for k in ProtoModelVersionDeploymentJobState.DeploymentJobRunState.keys()
    }
    _STATE_TO_STRING = {value: key for key, value in _STRING_TO_STATE.items()}

    @staticmethod
    def from_string(state_str):
        if state_str not in ModelVersionDeploymentJobRunState._STRING_TO_STATE:
            raise Exception(
                f"Could not get deployment job run state corresponding to string {state_str}. "
                f"Valid state strings: {ModelVersionDeploymentJobRunState.all_states()}"
            )
        return ModelVersionDeploymentJobRunState._STRING_TO_STATE[state_str]

    @staticmethod
    def to_string(state):
        if state not in ModelVersionDeploymentJobRunState._STATE_TO_STRING:
            raise Exception(
                f"Could not get string corresponding to deployment job run {state}. "
                f"Valid states: {ModelVersionDeploymentJobRunState.all_states()}"
            )
        return ModelVersionDeploymentJobRunState._STATE_TO_STRING[state]

    @staticmethod
    def all_states():
        return list(ModelVersionDeploymentJobRunState._STATE_TO_STRING.keys())



import logging

from mlflow.strands.autolog import setup_strands_tracing, teardown_strands_tracing
from mlflow.utils.annotations import experimental
from mlflow.utils.autologging_utils import autologging_integration

FLAVOR_NAME = "strands"
_logger = logging.getLogger(__name__)


@experimental(version="3.4.0")
def autolog(log_traces: bool = True, disable: bool = False, silent: bool = False):
    """
    Enables (or disables) and configures autologging from Strands Agents SDK to MLflow.

    Args:
        log_traces: If ``True``, traces are logged for Strands Agents.
        disable: If ``True``, disables Strands autologging.
        silent: If ``True``, suppresses all MLflow event logs and warnings.
    """
    _autolog(log_traces=log_traces, disable=disable, silent=silent)
    if disable or not log_traces:
        teardown_strands_tracing()
    else:
        setup_strands_tracing()


# This is required by mlflow.autolog()
autolog.integration_name = FLAVOR_NAME


@autologging_integration(FLAVOR_NAME)
def _autolog(log_traces: bool = True, disable: bool = False, silent: bool = False):
    """
    This function exists solely to attach the autologging_integration decorator without
    preventing cleanup logic from running when disable=True. Do not add implementation here.
    """



"""DSPy-based alignment optimizer implementation."""

import logging
from abc import abstractmethod
from typing import Any, Callable, ClassVar, Collection

from mlflow.entities.assessment import Feedback
from mlflow.entities.trace import Trace
from mlflow.exceptions import MlflowException
from mlflow.genai.judges import make_judge
from mlflow.genai.judges.base import AlignmentOptimizer, Judge
from mlflow.genai.judges.constants import _DATABRICKS_DEFAULT_JUDGE_MODEL
from mlflow.genai.judges.optimizers.dspy_utils import (
    agreement_metric,
    construct_dspy_lm,
    convert_litellm_to_mlflow_uri,
    create_dspy_signature,
    trace_to_dspy_example,
)
from mlflow.genai.judges.utils import _suppress_litellm_nonfatal_errors, get_default_model
from mlflow.protos.databricks_pb2 import INTERNAL_ERROR, INVALID_PARAMETER_VALUE
from mlflow.utils.annotations import experimental

# Import dspy - raise exception if not installed
try:
    import dspy
except ImportError:
    raise MlflowException(
        "DSPy library is required but not installed. Please install it with: pip install dspy",
        error_code=INTERNAL_ERROR,
    )

_logger = logging.getLogger(__name__)


@experimental(version="3.4.0")
class DSPyAlignmentOptimizer(AlignmentOptimizer):
    """
    Abstract base class for DSPy-based alignment optimizers.

    Provides common functionality for converting MLflow traces to DSPy examples
    and handling DSPy program compilation.
    """

    _logger: logging.Logger
    _model: str

    _MINIMUM_TRACES_REQUIRED_FOR_OPTIMIZATION: ClassVar[int] = 10

    @classmethod
    def get_min_traces_required(cls) -> int:
        """Get the minimum number of traces required for optimization.

        Returns:
            The minimum number of traces required for optimization.
        """
        return cls._MINIMUM_TRACES_REQUIRED_FOR_OPTIMIZATION

    @property
    def model(self) -> str:
        """Get the model used by this optimizer."""
        return self._model

    def __init__(self, model: str | None = None, **kwargs):
        """
        Initialize DSPy optimizer with common parameters.

        Args:
            model: Model to use for DSPy optimization. If None, uses get_default_model().
            **kwargs: Additional keyword arguments.
        """
        super().__init__(**kwargs)
        self._logger = logging.getLogger(self.__class__.__name__)
        self._model = model if model is not None else get_default_model()

    @abstractmethod
    def _dspy_optimize(
        self,
        program: "dspy.Module",
        examples: Collection["dspy.Example"],
        metric_fn: Callable[["dspy.Example", Any, Any | None], bool],
    ) -> "dspy.Module":
        """
        Perform DSPy optimization with algorithm-specific parameters.

        Each implementation can decide how to split the data internally if needed.

        Args:
            program: The DSPy program to optimize
            examples: Examples for optimization (implementations decide how to split)
            metric_fn: Metric function for optimization

        Returns:
            Optimized DSPy program
        """

    def _get_dspy_program_from_judge(self, judge: Judge) -> Any:
        """Convert a judge into a DSPy Predict module."""

        class CustomPredict(dspy.Predict):
            """
            Custom DSPy Predict class that allows passing an LM to the forward method.
            This is necessary to ensure that the optimized dspy program uses the judge's model,
            while we allow for the optimizer itself to use a different model.
            """

            def __init__(self, judge):
                super().__init__(create_dspy_signature(judge))
                self._judge_model: str = judge.model
                self._judge_name: str = judge.name

            def forward(self, *args, **kwargs):
                # If an LLM is supplied via kwargs, extract the model URI and use it,
                # else use self._judge_model
                dspy_lm: dspy.LM = kwargs.pop("lm", None)
                if dspy_lm is not None:
                    if dspy_lm.model == _DATABRICKS_DEFAULT_JUDGE_MODEL:
                        # The databricks default judge model is a special sentinel value
                        # and is not a valid LiteLLM model identifier
                        judge_model = _DATABRICKS_DEFAULT_JUDGE_MODEL
                    else:
                        judge_model = convert_litellm_to_mlflow_uri(dspy_lm.model)
                else:
                    judge_model = self._judge_model

                judge: Judge = make_judge(
                    name=self._judge_name,
                    instructions=self.signature.instructions,
                    model=judge_model,
                )
                feedback: Feedback = judge(**kwargs)
                return dspy.Prediction(
                    result=feedback.value,
                    rationale=feedback.rationale,
                )

        return CustomPredict(judge)

    @_suppress_litellm_nonfatal_errors
    def align(self, judge: Judge, traces: list[Trace]) -> Judge:
        """
        Main alignment method that orchestrates the DSPy optimization process.

        1. Extract judge instructions and create DSPy signature
        2. Convert traces to DSPy examples
        3. Create and compile DSPy optimizer
        4. Generate optimized judge from results

        Args:
            judge: The judge to be optimized
            traces: List of traces containing alignment data.
                   The implementation will split these traces internally for train/validation.

        Returns:
            A new optimized Judge instance
        """
        try:
            if not traces:
                raise MlflowException(
                    "No traces provided for alignment",
                    error_code=INVALID_PARAMETER_VALUE,
                )

            self._logger.debug(f"Setting up DSPy context with model: {self._model}")

            # Configure DSPy to use the optimizer's model
            # This ensures the optimizer uses its own model, separate from the judge's model
            optimizer_lm = construct_dspy_lm(self._model)

            with dspy.context(lm=optimizer_lm):
                # Create DSPy program that will simulate the judge
                program = self._get_dspy_program_from_judge(judge)
                self._logger.debug("Created DSPy program with signature using judge's model")

                # Convert traces to DSPy format
                dspy_examples = []
                for trace in traces:
                    example = trace_to_dspy_example(trace, judge)
                    if example is not None:
                        dspy_examples.append(example)

                self._logger.info(
                    f"Preparing optimization with {len(dspy_examples)} examples "
                    f"from {len(traces)} traces"
                )

                if not dspy_examples:
                    raise MlflowException(
                        f"No valid examples could be created from traces. "
                        f"Ensure that the provided traces contain Feedback entries "
                        f"with name {judge.name}",
                        error_code=INVALID_PARAMETER_VALUE,
                    )

                min_traces = self.get_min_traces_required()
                if len(dspy_examples) < min_traces:
                    raise MlflowException(
                        f"At least {min_traces} valid traces are required for optimization. "
                        f"Label more traces with Feedback entries with name {judge.name}",
                        error_code=INVALID_PARAMETER_VALUE,
                    )

                self._logger.debug("Starting DSPy optimization...")

                # Use the algorithm-specific optimization method
                # Each implementation decides how to handle data splitting
                optimized_program = self._dspy_optimize(program, dspy_examples, agreement_metric)

                self._logger.debug("DSPy optimization completed")

                # Create optimized judge with DSPy-optimized instructions

                optimized_instructions = optimized_program.signature.instructions
                return make_judge(
                    name=judge.name, instructions=optimized_instructions, model=judge.model
                )

        except Exception as e:
            raise MlflowException(
                f"Alignment optimization failed: {e!s}", error_code=INTERNAL_ERROR
            ) from e



import importlib
import inspect
import logging
import warnings
from typing import Any, Generator

from packaging import version

from mlflow.models.resources import (
    DatabricksFunction,
    DatabricksServingEndpoint,
    DatabricksSQLWarehouse,
    DatabricksVectorSearchIndex,
    Resource,
)

_logger = logging.getLogger(__name__)


def _get_embedding_model_endpoint_names(index):
    embedding_model_endpoint_names = []
    desc = index.describe()
    delta_sync_index_spec = desc.get("delta_sync_index_spec", {})
    embedding_source_columns = delta_sync_index_spec.get("embedding_source_columns", [])
    for column in embedding_source_columns:
        embedding_model_endpoint_name = column.get("embedding_model_endpoint_name", None)
        if embedding_model_endpoint_name:
            embedding_model_endpoint_names.append(embedding_model_endpoint_name)
    return embedding_model_endpoint_names


def _get_vectorstore_from_retriever(retriever) -> Generator[Resource, None, None]:
    vectorstore = getattr(retriever, "vectorstore", None)
    if _isinstance_with_multiple_modules(
        vectorstore,
        "DatabricksVectorSearch",
        [
            "databricks_langchain",
            "langchain_databricks",
            "langchain_community.vectorstores",
            "langchain.vectorstores",
        ],
    ):
        index = vectorstore.index
        yield DatabricksVectorSearchIndex(index_name=index.name)
        for embedding_endpoint in _get_embedding_model_endpoint_names(index):
            yield DatabricksServingEndpoint(endpoint_name=embedding_endpoint)

    embeddings = getattr(vectorstore, "embeddings", None)
    if _isinstance_with_multiple_modules(
        embeddings,
        "DatabricksEmbeddings",
        [
            "databricks_langchain",
            "langchain_databricks",
            "langchain_community.embeddings",
            "langchain.embeddings",
        ],
    ):
        yield DatabricksServingEndpoint(endpoint_name=embeddings.endpoint)


def _is_langchain_community_uc_function_toolkit(obj):
    try:
        from langchain_community.tools.databricks import UCFunctionToolkit
    except Exception:
        return False

    return isinstance(obj, UCFunctionToolkit)


def _is_unitycatalog_tool(obj):
    try:
        from unitycatalog.ai.langchain.toolkit import UnityCatalogTool
    except Exception:
        return False

    return isinstance(obj, UnityCatalogTool)


def _extract_databricks_dependencies_from_tools(tools) -> Generator[Resource, None, None]:
    if isinstance(tools, list):
        warehouse_ids = set()
        for tool in tools:
            if _isinstance_with_multiple_modules(
                tool, "BaseTool", ["langchain_core.tools", "langchain_community.tools"]
            ):
                # Handle Retriever tools
                if hasattr(tool.func, "keywords") and "retriever" in tool.func.keywords:
                    retriever = tool.func.keywords.get("retriever")
                    yield from _get_vectorstore_from_retriever(retriever)
                elif _is_unitycatalog_tool(tool):
                    if warehouse_id := tool.client_config.get("warehouse_id"):
                        warehouse_ids.add(warehouse_id)
                    yield DatabricksFunction(function_name=tool.uc_function_name)
                else:
                    # Tools here are a part of the BaseTool and have no attribute of a
                    # WarehouseID Extract the global variables of the function defined
                    # in the tool to get the UCFunctionToolkit Constants
                    nonlocal_vars = inspect.getclosurevars(tool.func).nonlocals
                    if "self" in nonlocal_vars and _is_langchain_community_uc_function_toolkit(
                        nonlocal_vars.get("self")
                    ):
                        uc_function_toolkit = nonlocal_vars.get("self")
                        # As we are iterating through each tool, adding a warehouse id everytime
                        # is a duplicative resource. Use a set to dedup warehouse ids and add
                        # them in the end
                        warehouse_ids.add(uc_function_toolkit.warehouse_id)

                        # In langchain the names of the tools are modified to have underscores:
                        # main.catalog.test_func -> main_catalog_test_func
                        # The original name of the tool is stored as the key in the tools
                        # dictionary. This code finds the correct tool and extract the key
                        langchain_tool_name = tool.name
                        filtered_tool_names = [
                            tool_name
                            for tool_name, uc_tool in uc_function_toolkit.tools.items()
                            if uc_tool.name == langchain_tool_name
                        ]
                        # This should always have the length 1
                        for tool_name in filtered_tool_names:
                            yield DatabricksFunction(function_name=tool_name)
        # Add the deduped warehouse ids
        for warehouse_id in warehouse_ids:
            yield DatabricksSQLWarehouse(warehouse_id=warehouse_id)


def _extract_databricks_dependencies_from_retriever(retriever) -> Generator[Resource, None, None]:
    # ContextualCompressionRetriever uses attribute "base_retriever"
    if hasattr(retriever, "base_retriever"):
        retriever = getattr(retriever, "base_retriever", None)

    # Most other retrievers use attribute "retriever"
    if hasattr(retriever, "retriever"):
        retriever = getattr(retriever, "retriever", None)

    # EnsembleRetriever uses attribute "retrievers" for multiple retrievers
    if hasattr(retriever, "retrievers"):
        retriever = getattr(retriever, "retrievers", None)

    # If there are multiple retrievers, we iterate over them to get dependencies from each of them
    if isinstance(retriever, list):
        for single_retriever in retriever:
            yield from _get_vectorstore_from_retriever(single_retriever)
    else:
        yield from _get_vectorstore_from_retriever(retriever)


def _extract_databricks_dependencies_from_llm(llm) -> Generator[Resource, None, None]:
    if _isinstance_with_multiple_modules(
        llm, "Databricks", ["langchain.llms", "langchain_community.llms"]
    ):
        yield DatabricksServingEndpoint(endpoint_name=llm.endpoint_name)


def _extract_databricks_dependencies_from_chat_model(chat_model) -> Generator[Resource, None, None]:
    if _isinstance_with_multiple_modules(
        chat_model,
        "ChatDatabricks",
        [
            "databricks_langchain",
            "langchain_databricks",
            "langchain.chat_models",
            "langchain_community.chat_models",
        ],
    ):
        yield DatabricksServingEndpoint(endpoint_name=chat_model.endpoint)


def _extract_databricks_dependencies_from_tool_nodes(tool_node) -> Generator[Resource, None, None]:
    try:
        try:
            # LangGraph >= 0.3
            from langgraph.prebuilt import ToolNode
        except ImportError:
            # LangGraph < 0.3
            from langgraph.prebuilt.tool_node import ToolNode

        if isinstance(tool_node, ToolNode):
            yield from _extract_databricks_dependencies_from_tools(
                list(tool_node.tools_by_name.values())
            )
    except ImportError:
        pass


def _isinstance_with_multiple_modules(
    object: Any, class_name: str, from_modules: list[str]
) -> bool:
    """
    Databricks components are defined in different modules in LangChain e.g.
    langchain, langchain_community, databricks_langchain due to historical migrations.
    To keep backward compatibility, we need to check if the object is an instance of the
    class defined in any of those different modules.

    Args:
        object: The object to check
        class_name: The name of the class to check
        from_modules: The list of modules to import the class from.
    """
    # Suppress LangChainDeprecationWarning for old imports
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", DeprecationWarning)

        for module_path in from_modules:
            try:
                module = importlib.import_module(module_path)
                cls = getattr(module, class_name)

                if cls is not None and isinstance(object, cls):
                    return True
            except (ImportError, AttributeError):
                pass

    return False


_LEGACY_MODEL_ATTR_SET = {
    "llm",  # LLMChain
    "retriever",  # RetrievalQA
    "llm_chain",  # StuffDocumentsChain, MapRerankDocumentsChain, MapReduceDocumentsChain
    "question_generator",  # BaseConversationalRetrievalChain
    "initial_llm_chain",  # RefineDocumentsChain
    "refine_llm_chain",  # RefineDocumentsChain
    "combine_documents_chain",  # RetrievalQA, ReduceDocumentsChain
    "combine_docs_chain",  # BaseConversationalRetrievalChain
    "collapse_documents_chain",  # ReduceDocumentsChain,
    "agent",  # Agent,
    "tools",  # Tools
}


def _extract_dependency_list_from_lc_model(lc_model) -> Generator[Resource, None, None]:
    """
    This function contains the logic to examine a non-Runnable component of a langchain model.
    The logic here does not cover all legacy chains. If you need to support a custom chain,
    you need to monkey patch this function.
    """
    if lc_model is None:
        return

    # leaf node
    yield from _extract_databricks_dependencies_from_chat_model(lc_model)
    yield from _extract_databricks_dependencies_from_retriever(lc_model)
    yield from _extract_databricks_dependencies_from_llm(lc_model)
    yield from _extract_databricks_dependencies_from_tools(lc_model)
    yield from _extract_databricks_dependencies_from_tool_nodes(lc_model)

    # recursively inspect legacy chain
    for attr_name in _LEGACY_MODEL_ATTR_SET:
        yield from _extract_dependency_list_from_lc_model(getattr(lc_model, attr_name, None))


def _traverse_runnable(
    lc_model,
    visited: set[int] | None = None,
) -> Generator[Resource, None, None]:
    """
    This function contains the logic to traverse a langchain_core.runnables.RunnableSerializable
    object. It first inspects the current object using _extract_dependency_list_from_lc_model
    and then, if the current object is a Runnable, it recursively inspects its children returned
    by lc_model.get_graph().nodes.values().
    This function supports arbitrary LCEL chain.
    """
    import pydantic
    from langchain_core.runnables import Runnable, RunnableLambda

    visited = visited or set()
    current_object_id = id(lc_model)
    if current_object_id in visited:
        return

    # Visit the current object
    visited.add(current_object_id)
    yield from _extract_dependency_list_from_lc_model(lc_model)

    if isinstance(lc_model, Runnable):
        # Visit the returned graph
        if isinstance(lc_model, RunnableLambda) and version.parse(
            pydantic.version.VERSION
        ) >= version.parse("2.0"):
            nodes = _get_nodes_from_runnable_lambda(lc_model)
        else:
            nodes = _get_nodes_from_runnable_callable(lc_model)
            # If no nodes are found continue with the default behaviour
            if len(nodes) == 0:
                nodes = lc_model.get_graph().nodes.values()

        for node in nodes:
            yield from _traverse_runnable(node.data, visited)
    else:
        # No-op for non-runnable, if any
        pass


def _get_deps_from_closures(lc_model):
    """
    In some cases, the dependency extraction of Runnable Lambda fails because the call
    `inspect.getsource(func)` can fail. This causes deps of RunnableLambda to be empty.
    Therefore this method adds an additional way of getting dependencies through
    closure variables.

    TODO: Remove when issue gets resolved: https://github.com/langchain-ai/langchain/issues/27970
    """
    if not hasattr(lc_model, "func"):
        return []

    try:
        from langchain_core.runnables import Runnable

        closure = inspect.getclosurevars(lc_model.func)
        candidates = {**closure.globals, **closure.nonlocals}
        deps = []

        # This code is taken from Langchain deps here: https://github.com/langchain-ai/langchain/blob/14f182795312f01985344576b5199681683641e1/libs/core/langchain_core/runnables/base.py#L4481
        for _, v in candidates.items():
            if isinstance(v, Runnable):
                deps.append(v)
            elif isinstance(getattr(v, "__self__", None), Runnable):
                deps.append(v.__self__)

        return deps
    except Exception:
        return []


def _get_nodes_from_runnable_lambda(lc_model):
    """
    This is a workaround for the LangGraph issue: https://github.com/langchain-ai/langgraph/issues/1856

    For RunnableLambda, we calling lc_model.get_graph() to get the nodes, which inspect
    the input and output schema using wrapped function's type annotation. However, the
    prebuilt graph (e.g. create_react_agent) from LangGraph uses typing.TypeDict annotation,
    which is not supported by Pydantic V2 on Python < 3.12. If we try to inspect such
    function, it will raise the following error:

        pydantic.errors.PydanticUserError: Please use `typing_extensions.TypedDict`
        instead of`typing.TypedDict` on Python < 3.12. For further information visit
        https://errors.pydantic.dev/2.9/u/typed-dict-version

    Therefore, we cannot use get_graph() for RunnableLambda until LangGraph fixes this issue.
    Luckily, we are not interested in the input/output nodes for extracting databricks
    dependencies. We only care about lc_models.deps, which contains the components that
    the RunnableLambda depends on. Therefore, this function extracts the necessary parts
    from the original get_graph() function, dropping the input/output related logic.
    https://github.com/langchain-ai/langchain/blob/2ea5f60cc5747a334550273a5dba1b70b11414c1/libs/core/langchain_core/runnables/base.py#L4493C1-L4512C46
    """

    if deps := lc_model.deps or _get_deps_from_closures(lc_model):
        nodes = []
        for dep in deps:
            dep_graph = dep.get_graph()
            dep_graph.trim_first_node()
            dep_graph.trim_last_node()
            nodes.extend(dep_graph.nodes.values())
    else:
        nodes = lc_model.get_graph().nodes.values()
    return nodes


def _get_nodes_from_runnable_callable(lc_model):
    """
    RunnableLambda has a `deps` property which goes through the function and extracts a
    ny dependencies. RunnableCallable does not have this property so we cannot derive all
    the dependencies from the function. This helper method also looks into the function of the
    callable to retrieve these dependencies.

    The code here is from: https://github.com/langchain-ai/langchain/blob/12fea5b868edd12b0d576e7f8bfc922d0167eeab/libs/core/langchain_core/runnables/base.py#L4467
    """

    # If Runnable Callable is not importable or if the lc_model is not an instance
    # of RunnableCallable return early
    try:
        from langchain_core.runnables import Runnable
        from langchain_core.runnables.utils import get_function_nonlocals
        from langgraph.utils.runnable import RunnableCallable

        if not isinstance(lc_model, RunnableCallable):
            return []
    except ImportError:
        return []

    if hasattr(lc_model, "func"):
        objects = get_function_nonlocals(lc_model.func)
    elif hasattr(lc_model, "afunc"):
        objects = get_function_nonlocals(lc_model.afunc)
    else:
        objects = []

    deps = []
    for obj in objects:
        if isinstance(obj, Runnable):
            deps.append(obj)
        elif isinstance(getattr(obj, "__self__", None), Runnable):
            deps.append(obj.__self__)

    nodes = []
    for dep in deps:
        dep_graph = dep.get_graph()
        dep_graph.trim_first_node()
        dep_graph.trim_last_node()
        nodes.extend(dep_graph.nodes.values())
    return nodes


def _detect_databricks_dependencies(lc_model, log_errors_as_warnings=True) -> list[Resource]:
    """
    Detects the databricks dependencies of a langchain model and returns a list of
    detected endpoint names and index names.

    lc_model can be an arbitrary `chain that is built with LCEL <https://python.langchain.com/docs/modules/chains#lcel-chains>`_,
    which is a langchain_core.runnables.RunnableSerializable.
    `Legacy chains <https://python.langchain.com/docs/modules/chains#legacy-chains>`_ have limited
    support. Only RetrievalQA, StuffDocumentsChain, ReduceDocumentsChain, RefineDocumentsChain,
    MapRerankDocumentsChain, MapReduceDocumentsChain, BaseConversationalRetrievalChain are
    supported. If you need to support a custom chain, you need to monkey patch
    the function mlflow.langchain.databricks_dependencies._extract_dependency_list_from_lc_model().

    For an LCEL chain, all the langchain_core.runnables.RunnableSerializable nodes will be
    traversed.

    If a retriever is found, it will be used to extract the databricks vector search and embeddings
    dependencies.
    If an llm is found, it will be used to extract the databricks llm dependencies.
    If a chat_model is found, it will be used to extract the databricks chat dependencies.
    """
    try:
        dependency_list = list(_traverse_runnable(lc_model))
        # Filter out duplicate dependencies so same dependencies are not added multiple times
        # We can't use set here as the object is not hashable so we need to filter it out manually.
        unique_dependencies = []
        for dependency in dependency_list:
            if dependency not in unique_dependencies:
                unique_dependencies.append(dependency)
        return unique_dependencies
    except Exception:
        if log_errors_as_warnings:
            _logger.warning(
                "Unable to detect Databricks dependencies. "
                "Set logging level to DEBUG to see the full traceback."
            )
            _logger.debug("", exc_info=True)
            return []
        raise



"""
Shared types for MLflow GenAI judge tools.

This module provides common data structures and types that can be reused
across multiple judge tools for consistent data representation.
"""

from dataclasses import dataclass

from mlflow.entities.span_status import SpanStatus
from mlflow.utils.annotations import experimental


@experimental(version="3.4.0")
@dataclass
class SpanResult:
    """Result from getting span content."""

    span_id: str | None
    content: str | None
    content_size_bytes: int
    page_token: str | None = None
    error: str | None = None


@experimental(version="3.4.0")
@dataclass
class SpanInfo:
    """Information about a single span."""

    span_id: str
    name: str
    span_type: str
    start_time_ms: float
    end_time_ms: float
    duration_ms: float
    parent_id: str | None
    status: SpanStatus
    is_root: bool
    attribute_names: list[str]



import logging
from dataclasses import dataclass
from typing import Any

from mlflow.utils.autologging_utils import AUTOLOGGING_INTEGRATIONS

_logger = logging.getLogger(__name__)


@dataclass
class AutoLoggingConfig:
    """
    A dataclass to hold common autologging configuration options.
    """

    log_input_examples: bool
    log_model_signatures: bool
    log_traces: bool
    extra_tags: dict[str, Any] | None = None
    log_models: bool = True

    @classmethod
    def init(cls, flavor_name: str):
        config_dict = AUTOLOGGING_INTEGRATIONS.get(flavor_name, {})
        # NB: These defaults are only used when the autolog() function for the
        # flavor does not specify the corresponding configuration option
        return cls(
            log_models=config_dict.get("log_models", False),
            log_input_examples=config_dict.get("log_input_examples", False),
            log_model_signatures=config_dict.get("log_model_signatures", False),
            log_traces=config_dict.get("log_traces", True),
            extra_tags=config_dict.get("extra_tags", None),
        )



import hashlib
from typing import Any

from packaging.version import Version

from mlflow.exceptions import MlflowException
from mlflow.protos.databricks_pb2 import INVALID_PARAMETER_VALUE

MAX_ROWS = 10000


def compute_pandas_digest(df) -> str:
    """Computes a digest for the given Pandas DataFrame.

    Args:
        df: A Pandas DataFrame.

    Returns:
        A string digest.
    """
    import numpy as np
    import pandas as pd

    # trim to max rows
    trimmed_df = df.head(MAX_ROWS)

    # keep string and number columns, drop other column types
    if Version(pd.__version__) >= Version("2.1.0"):
        string_columns = trimmed_df.columns[(df.map(type) == str).all(0)]
    else:
        string_columns = trimmed_df.columns[(df.applymap(type) == str).all(0)]
    numeric_columns = trimmed_df.select_dtypes(include=[np.number]).columns

    desired_columns = string_columns.union(numeric_columns)
    trimmed_df = trimmed_df[desired_columns]

    return get_normalized_md5_digest(
        [
            pd.util.hash_pandas_object(trimmed_df).values,
            np.int64(len(df)),
        ]
        + [str(x).encode() for x in df.columns]
    )


def compute_numpy_digest(features, targets=None) -> str:
    """Computes a digest for the given numpy array.

    Args:
        features: A numpy array containing dataset features.
        targets: A numpy array containing dataset targets. Optional.

    Returns:
        A string digest.
    """
    import numpy as np
    import pandas as pd

    hashable_elements = []

    def hash_array(array):
        flattened_array = array.flatten()
        trimmed_array = flattened_array[0:MAX_ROWS]
        try:
            hashable_elements.append(pd.util.hash_array(trimmed_array))
        except TypeError:
            hashable_elements.append(np.int64(trimmed_array.size))

        # hash full array dimensions
        for x in array.shape:
            hashable_elements.append(np.int64(x))

    def hash_dict_of_arrays(array_dict):
        for key in sorted(array_dict.keys()):
            hash_array(array_dict[key])

    for item in [features, targets]:
        if item is None:
            continue
        if isinstance(item, dict):
            hash_dict_of_arrays(item)
        else:
            hash_array(item)

    return get_normalized_md5_digest(hashable_elements)


def get_normalized_md5_digest(elements: list[Any]) -> str:
    """Computes a normalized digest for a list of hashable elements.

    Args:
        elements: A list of hashable elements for inclusion in the md5 digest.

    Returns:
        An 8-character, truncated md5 digest.
    """

    if not elements:
        raise MlflowException(
            "No hashable elements were provided for md5 digest creation",
            INVALID_PARAMETER_VALUE,
        )

    md5 = hashlib.md5(usedforsecurity=False)
    for element in elements:
        md5.update(element)

    return md5.hexdigest()[:8]



import logging
from typing import Callable

from mlflow.utils.autologging_utils import autologging_integration, safe_patch

FLAVOR_NAME = "litellm"

_logger = logging.getLogger(__name__)


def autolog(
    log_traces: bool = True,
    disable: bool = False,
    silent: bool = False,
):
    """
    Enables (or disables) and configures autologging from LiteLLM to MLflow. Currently, MLflow
    only supports autologging for tracing.

    Args:
        log_traces: If ``True``, traces are logged for LiteLLM calls. If ``False``,
            no traces are collected during inference. Default to ``True``.
        disable: If ``True``, disables the LiteLLM autologging integration. If ``False``,
            enables the LiteLLM autologging integration.
        silent: If ``True``, suppress all event logs and warnings from MLflow during LiteLLM
            autologging. If ``False``, show all events and warnings.
    """
    import litellm

    # This needs to be called before doing any safe-patching (otherwise safe-patch will be no-op).
    # TODO: since this implementation is inconsistent, explore a universal way to solve the issue.
    _autolog(log_traces=log_traces, disable=disable, silent=silent)

    try:
        from litellm.integrations.mlflow import MlflowLogger  # noqa: F401
    except ImportError:
        _logger.warning(
            "MLflow LiteLLM integration is not supported for the installed LiteLLM version. "
            "Please upgrade to a newer version to enable MLflow LiteLLM autologging."
        )
        return

    if log_traces and not disable:
        litellm.success_callback = _append_mlflow_callbacks(litellm.success_callback)
        litellm.failure_callback = _append_mlflow_callbacks(litellm.failure_callback)

        # Patch thread pool executor to bypass non-blocking behavior of success_handler
        _patch_thread_pool()

    else:
        litellm.success_callback = _remove_mlflow_callbacks(litellm.success_callback)
        litellm.failure_callback = _remove_mlflow_callbacks(litellm.failure_callback)
        # Callback also needs to be removed from 'callbacks' as litellm adds
        # success/failure callbacks to there as well.
        litellm.callbacks = _remove_mlflow_callbacks(litellm.callbacks)


# This is required by mlflow.autolog()
autolog.integration_name = FLAVOR_NAME


# NB: The @autologging_integration annotation must be applied here, and the callback injection
# needs to happen outside the annotated function. This is because the annotated function is NOT
# executed when disable=True is passed. This prevents us from removing our callback and patching
# when autologging is turned off.
@autologging_integration(FLAVOR_NAME)
def _autolog(
    log_traces: bool,
    disable: bool = False,
    silent: bool = False,
):
    pass


def _patch_thread_pool():
    """
    Apply the threading patch to a synchronous function.

    We capture the threads started by the function using the _patch_thread_start context manager,
    then join them to ensure they are finished before the notebook cell finishes executing.
    """
    try:
        from litellm.litellm_core_utils.thread_pool_executor import executor
    except ImportError:
        _logger.warning(
            "MLflow LiteLLM integration is not supported for the installed LiteLLM version. "
            "The behavior might be unstable."
        )
        return

    def _patched_submit(original, *args, **kwargs):
        if isinstance(args[0], Callable) and "success_handler" in args[0].__name__:
            # Immediately run the callback handler instead of submitting it to the thread pool
            args[0](*args[1:], **kwargs)
            return
        return original(*args, **kwargs)

    safe_patch(FLAVOR_NAME, executor, "submit", _patched_submit)


def _append_mlflow_callbacks(callbacks):
    from litellm.integrations.mlflow import MlflowLogger

    # MLflow callback can be stored as a string or the actual logger object
    if not any(cb == "mlflow" or isinstance(cb, MlflowLogger) for cb in callbacks):
        return callbacks + ["mlflow"]

    return callbacks


def _remove_mlflow_callbacks(callbacks):
    from litellm.integrations.mlflow import MlflowLogger

    return [cb for cb in callbacks if not (cb == "mlflow" or isinstance(cb, MlflowLogger))]



"""
OpenTelemetry REST API endpoints for MLflow FastAPI server.

This module implements the OpenTelemetry Protocol (OTLP) REST API for ingesting spans
according to the OTel specification:
https://opentelemetry.io/docs/specs/otlp/#otlphttp

Note: This is a minimal implementation that serves as a placeholder for the OTel endpoint.
The actual span ingestion logic would need to properly convert incoming OTel format spans
to MLflow spans, which requires more complex conversion logic.
"""

from typing import Any

from fastapi import APIRouter, Header, HTTPException, Request, Response, status
from google.protobuf.message import DecodeError
from opentelemetry.proto.collector.trace.v1.trace_service_pb2 import ExportTraceServiceRequest
from pydantic import BaseModel, Field

from mlflow.entities.span import Span
from mlflow.server.handlers import _get_tracking_store
from mlflow.tracing.utils.otlp import MLFLOW_EXPERIMENT_ID_HEADER, OTLP_TRACES_PATH

# Create FastAPI router for OTel endpoints
otel_router = APIRouter(prefix=OTLP_TRACES_PATH, tags=["OpenTelemetry"])


class OTelExportTraceServiceResponse(BaseModel):
    """
    Pydantic model for the OTLP/HTTP ExportTraceServiceResponse.

    This matches the OpenTelemetry protocol specification for trace export responses.
    Reference: https://opentelemetry.io/docs/specs/otlp/
    """

    partialSuccess: dict[str, Any] | None = Field(
        None, description="Details about partial success of the export operation"
    )


@otel_router.post("", response_model=OTelExportTraceServiceResponse, status_code=200)
async def export_traces(
    request: Request,
    response: Response,
    x_mlflow_experiment_id: str = Header(..., alias=MLFLOW_EXPERIMENT_ID_HEADER),
    content_type: str = Header(None),
) -> OTelExportTraceServiceResponse:
    """
    Export trace spans to MLflow via the OpenTelemetry protocol.

    This endpoint accepts OTLP/HTTP protobuf trace export requests.
    Protobuf format reference: https://opentelemetry.io/docs/specs/otlp/#binary-protobuf-encoding

    Args:
        request: OTel ExportTraceServiceRequest in protobuf format
        response: FastAPI Response object for setting headers
        x_mlflow_experiment_id: Required header containing the experiment ID
        content_type: Content-Type header from the request

    Returns:
        OTel ExportTraceServiceResponse indicating success

    Raises:
        HTTPException: If the request is invalid or span logging fails
    """
    # Validate Content-Type header
    if content_type != "application/x-protobuf":
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=f"Invalid Content-Type: {content_type}. Expected: application/x-protobuf",
        )

    # Set response Content-Type header
    response.headers["Content-Type"] = "application/x-protobuf"

    body = await request.body()
    parsed_request = ExportTraceServiceRequest()

    try:
        # In Python protobuf library 5.x, ParseFromString may not raise DecodeError on invalid data
        parsed_request.ParseFromString(body)

        # Check if we actually parsed any data
        # If no resource_spans were parsed, the data was likely invalid
        if not parsed_request.resource_spans:
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST,
                detail="Invalid OpenTelemetry protobuf format - no spans found",
            )

    except DecodeError:
        # This will catch errors in Python protobuf library 3.x
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail="Invalid OpenTelemetry protobuf format",
        )

    mlflow_spans = []
    for resource_span in parsed_request.resource_spans:
        for scope_span in resource_span.scope_spans:
            for otel_proto_span in scope_span.spans:
                try:
                    mlflow_span = Span._from_otel_proto(otel_proto_span)
                    mlflow_spans.append(mlflow_span)
                except Exception:
                    raise HTTPException(
                        status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
                        detail="Cannot convert OpenTelemetry span to MLflow span",
                    )

    if mlflow_spans:
        store = _get_tracking_store()

        try:
            store.log_spans(x_mlflow_experiment_id, mlflow_spans)
        except NotImplementedError:
            raise HTTPException(
                status_code=status.HTTP_501_NOT_IMPLEMENTED,
                detail=f"REST OTLP span logging is not supported by {store.__class__.__name__}",
            )
        except Exception as e:
            raise HTTPException(
                status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
                detail=f"Cannot store OpenTelemetry spans: {e}",
            )

    return OTelExportTraceServiceResponse()



from typing import Any

from mlflow.entities.assessment import (
    DEFAULT_FEEDBACK_NAME,
    Assessment,
    AssessmentError,
    Expectation,
    Feedback,
    FeedbackValueType,
)
from mlflow.entities.assessment_source import AssessmentSource
from mlflow.exceptions import MlflowException
from mlflow.tracing.client import TracingClient
from mlflow.utils.annotations import experimental


@experimental(version="3.0.0")
def get_assessment(trace_id: str, assessment_id: str) -> Assessment:
    """
    Get an assessment entity from the backend store.

    Args:
        trace_id: The ID of the trace.
        assessment_id: The ID of the assessment to get.

    Returns:
        :py:class:`~mlflow.entities.Assessment`: The Assessment object.
    """
    return TracingClient().get_assessment(trace_id, assessment_id)


def log_assessment(trace_id: str, assessment: Assessment) -> Assessment:
    """
    Logs an assessment to a Trace. The assessment can be an expectation or a feedback.

    - Expectation: A label that represents the expected value for a particular operation.
        For example, an expected answer for a user question from a chatbot.
    - Feedback: A label that represents the feedback on the quality of the operation.
        Feedback can come from different sources, such as human judges, heuristic scorers,
        or LLM-as-a-Judge.

    The following code annotates a trace with a feedback provided by LLM-as-a-Judge.

    .. code-block:: python

        import mlflow
        from mlflow.entities import Feedback

        feedback = Feedback(
            name="faithfulness",
            value=0.9,
            rationale="The model is faithful to the input.",
            metadata={"model": "gpt-4o-mini"},
        )

        mlflow.log_assessment(trace_id="1234", assessment=feedback)

    The following code annotates a trace with human-provided ground truth with source information.
    When the source is not provided, the default source is set to "default" with type "HUMAN"

    .. code-block:: python

        import mlflow
        from mlflow.entities import AssessmentSource, AssessmentSourceType, Expectation

        # Specify the annotator information as a source.
        source = AssessmentSource(
            source_type=AssessmentSourceType.HUMAN,
            source_id="john@example.com",
        )

        expectation = Expectation(
            name="expected_answer",
            value=42,
            source=source,
        )

        mlflow.log_assessment(trace_id="1234", assessment=expectation)

    The expectation value can be any JSON-serializable value. For example, you may
     record the full LLM message as the expectation value.

    .. code-block:: python

        import mlflow
        from mlflow.entities.assessment import Expectation

        expectation = Expectation(
            name="expected_message",
            # Full LLM message including expected tool calls
            value={
                "role": "assistant",
                "content": "The answer is 42.",
                "tool_calls": [
                    {
                        "id": "1234",
                        "type": "function",
                        "function": {"name": "add", "arguments": "40 + 2"},
                    }
                ],
            },
        )
        mlflow.log_assessment(trace_id="1234", assessment=expectation)

    You can also log an error information during the feedback generation process. To do so,
    provide an instance of :py:class:`~mlflow.entities.AssessmentError` to the `error`
    parameter, and leave the `value` parameter as `None`.

    .. code-block:: python

        import mlflow
        from mlflow.entities import AssessmentError, Feedback

        error = AssessmentError(
            error_code="RATE_LIMIT_EXCEEDED",
            error_message="Rate limit for the judge exceeded.",
        )

        feedback = Feedback(
            trace_id="1234",
            name="faithfulness",
            error=error,
        )
        mlflow.log_assessment(trace_id="1234", assessment=feedback)


    """
    return TracingClient().log_assessment(trace_id, assessment)


@experimental(version="3.0.0")
def log_expectation(
    *,
    trace_id: str,
    name: str,
    value: Any,
    source: AssessmentSource | None = None,
    metadata: dict[str, Any] | None = None,
    span_id: str | None = None,
) -> Assessment:
    """
    Logs an expectation (e.g. ground truth label) to a Trace. This API only takes keyword arguments.

    Args:
        trace_id: The ID of the trace.
        name: The name of the expectation assessment e.g., "expected_answer
        value: The value of the expectation. It can be any JSON-serializable value.
        source: The source of the expectation assessment. Must be an instance of
                :py:class:`~mlflow.entities.AssessmentSource`. If not provided,
                default to CODE source type.
        metadata: Additional metadata for the expectation.
        span_id: The ID of the span associated with the expectation, if it needs be
                associated with a specific span in the trace.

    Returns:
        :py:class:`~mlflow.entities.Assessment`: The created expectation assessment.

    Examples:
        .. code-block:: python

            import mlflow
            from mlflow.entities import AssessmentSource, AssessmentSourceType

            # Log simple expected answer
            expectation = mlflow.log_expectation(
                trace_id="tr-1234567890abcdef",
                name="expected_answer",
                value="The capital of France is Paris.",
                source=AssessmentSource(
                    source_type=AssessmentSourceType.HUMAN, source_id="annotator@company.com"
                ),
                metadata={"question_type": "factual", "difficulty": "easy"},
            )

            # Log expected classification label
            mlflow.log_expectation(
                trace_id="tr-1234567890abcdef",
                name="expected_category",
                value="positive",
                source=AssessmentSource(
                    source_type=AssessmentSourceType.HUMAN, source_id="data_labeler_001"
                ),
                metadata={"labeling_session": "batch_01", "confidence": 0.95},
            )
    """
    assessment = Expectation(
        name=name,
        source=source,
        value=value,
        metadata=metadata,
        span_id=span_id,
    )
    return TracingClient().log_assessment(trace_id, assessment)


def update_assessment(
    trace_id: str,
    assessment_id: str,
    assessment: Assessment,
) -> Assessment:
    """
    Updates an existing expectation (ground truth) in a Trace.

    Args:
        trace_id: The ID of the trace.
        assessment_id: The ID of the expectation or feedback assessment to update.
        assessment: The updated assessment.

    Returns:
        :py:class:`~mlflow.entities.Assessment`: The updated feedback or expectation assessment.

    Example:

    The following code updates an existing expectation with a new value.
    To update other fields, provide the corresponding parameters.

    .. code-block:: python

        import mlflow
        from mlflow.entities import Expectation, ExpectationValue

        # Create an expectation with value 42.
        response = mlflow.log_assessment(
            trace_id="1234",
            assessment=Expectation(name="expected_answer", value=42),
        )
        assessment_id = response.assessment_id

        # Update the expectation with a new value 43.
        mlflow.update_assessment(
            trace_id="1234",
            assessment_id=assessment.assessment_id,
            assessment=Expectation(name="expected_answer", value=43),
        )
    """
    return TracingClient().update_assessment(
        assessment_id=assessment_id,
        trace_id=trace_id,
        assessment=assessment,
    )


def delete_assessment(trace_id: str, assessment_id: str):
    """
    Deletes an assessment associated with a trace.

    Args:
        trace_id: The ID of the trace.
        assessment_id: The ID of the assessment to delete.
    """
    return TracingClient().delete_assessment(trace_id=trace_id, assessment_id=assessment_id)


def log_feedback(
    *,
    trace_id: str,
    name: str = DEFAULT_FEEDBACK_NAME,
    value: FeedbackValueType | None = None,
    source: AssessmentSource | None = None,
    error: Exception | AssessmentError | None = None,
    rationale: str | None = None,
    metadata: dict[str, Any] | None = None,
    span_id: str | None = None,
) -> Assessment:
    """
    Logs feedback to a Trace. This API only takes keyword arguments.

    Args:
        trace_id: The ID of the trace.
        name: The name of the feedback assessment e.g., "faithfulness". Defaults to
            "feedback" if not provided.
        value: The value of the feedback. Must be one of the following types:
            - float
            - int
            - str
            - bool
            - list of values of the same types as above
            - dict with string keys and values of the same types as above
        source: The source of the feedback assessment. Must be an instance of
                :py:class:`~mlflow.entities.AssessmentSource`. If not provided, defaults to
                CODE source type
        error: An error object representing any issues encountered while computing the
            feedback, e.g., a timeout error from an LLM judge. Accepts an exception
            object, or an :py:class:`~mlflow.entities.AssessmentError` object. Either
            this or `value` must be provided.
        rationale: The rationale / justification for the feedback.
        metadata: Additional metadata for the feedback.
        span_id: The ID of the span associated with the feedback, if it needs be
                associated with a specific span in the trace.

    Returns:
        :py:class:`~mlflow.entities.Assessment`: The created feedback assessment.

    Examples:
        .. code-block:: python

            import mlflow
            from mlflow.entities import AssessmentSource, AssessmentSourceType

            # Log simple feedback score
            feedback = mlflow.log_feedback(
                trace_id="tr-1234567890abcdef",
                name="relevance",
                value=0.9,
                source=AssessmentSource(source_type=AssessmentSourceType.LLM, source_id="gpt-4"),
                rationale="Response directly addresses the user's question",
            )

            # Log detailed feedback with structured data
            mlflow.log_feedback(
                trace_id="tr-1234567890abcdef",
                name="quality_metrics",
                value={"accuracy": 0.95, "completeness": 0.88, "clarity": 0.92, "overall": 0.92},
                source=AssessmentSource(
                    source_type=AssessmentSourceType.HUMAN, source_id="expert_evaluator"
                ),
                rationale="High accuracy and clarity, slightly incomplete coverage",
            )
    """
    assessment = Feedback(
        name=name,
        source=source,
        value=value,
        error=error,
        rationale=rationale,
        metadata=metadata,
        span_id=span_id,
    )
    return TracingClient().log_assessment(trace_id, assessment)


@experimental(version="3.0.0")
def override_feedback(
    *,
    trace_id: str,
    assessment_id: str,
    value: FeedbackValueType,
    rationale: str | None = None,
    source: AssessmentSource | None = None,
    metadata: dict[str, Any] | None = None,
) -> Assessment:
    """
    Overrides an existing feedback assessment with a new assessment. This API
    logs a new assessment with the `overrides` field set to the provided assessment ID.
    The original assessment will be marked as invalid, but will otherwise be unchanged.
    This is useful when you want to correct an assessment generated by an LLM judge,
    but want to preserve the original assessment for future judge fine-tuning.

    If you want to mutate an assessment in-place, use :py:func:`update_assessment` instead.

    Args:
        trace_id: The ID of the trace.
        assessment_id: The ID of the assessment to override.
        value: The new value of the assessment.
        rationale: The rationale of the new assessment.
        source: The source of the new assessment.
        metadata: Additional metadata for the new assessment.

    Returns:
        :py:class:`~mlflow.entities.Assessment`: The created assessment.

    Examples:
        .. code-block:: python

            import mlflow
            from mlflow.entities import AssessmentSource, AssessmentSourceType

            # First, log an initial LLM-generated feedback as a simulation
            llm_feedback = mlflow.log_feedback(
                trace_id="tr-1234567890abcdef",
                name="relevance",
                value=0.6,
                source=AssessmentSource(source_type=AssessmentSourceType.LLM, source_id="gpt-4"),
                rationale="Response partially addresses the question",
            )

            # Later, a human reviewer disagrees and wants to override
            corrected_assessment = mlflow.override_feedback(
                trace_id="tr-1234567890abcdef",
                assessment_id=llm_feedback.assessment_id,
                value=0.9,
                rationale="Response fully addresses the question with good examples",
                source=AssessmentSource(
                    source_type=AssessmentSourceType.HUMAN, source_id="expert_reviewer@company.com"
                ),
                metadata={
                    "override_reason": "LLM underestimated relevance",
                    "review_date": "2024-01-15",
                    "confidence": "high",
                },
            )
    """
    old_assessment = get_assessment(trace_id, assessment_id)
    if not isinstance(old_assessment, Feedback):
        raise MlflowException.invalid_parameter_value(
            f"The assessment with ID {assessment_id} is not a feedback assessment."
        )

    new_assessment = Feedback(
        name=old_assessment.name,
        span_id=old_assessment.span_id,
        value=value,
        rationale=rationale,
        source=source,
        metadata=metadata,
        overrides=old_assessment.assessment_id,
    )

    return TracingClient().log_assessment(trace_id, new_assessment)



from mlflow.genai.optimize.optimizers.base_optimizer import BasePromptOptimizer
from mlflow.genai.optimize.optimizers.dspy_mipro_optimizer import _DSPyMIPROv2Optimizer
from mlflow.genai.optimize.optimizers.dspy_optimizer import DSPyPromptOptimizer

__all__ = ["BasePromptOptimizer", "DSPyPromptOptimizer", "_DSPyMIPROv2Optimizer"]



import json
import logging

import numpy as np

from mlflow.environment_variables import MLFLOW_INPUT_EXAMPLE_INFERENCE_TIMEOUT
from mlflow.models.signature import ModelSignature, infer_signature
from mlflow.models.utils import _contains_params
from mlflow.types.schema import ColSpec, DataType, Schema, TensorSpec
from mlflow.utils.os import is_windows
from mlflow.utils.timeout import MlflowTimeoutError, run_with_timeout

_logger = logging.getLogger(__name__)


_TEXT2TEXT_SIGNATURE = ModelSignature(
    inputs=Schema([ColSpec("string")]),
    outputs=Schema([ColSpec("string")]),
)
_CLASSIFICATION_SIGNATURE = ModelSignature(
    inputs=Schema([ColSpec("string")]),
    outputs=Schema([ColSpec("string", name="label"), ColSpec("double", name="score")]),
)

# Order is important here, the first matching task type will be used
_DEFAULT_SIGNATURE_FOR_TASK = {
    "token-classification": _TEXT2TEXT_SIGNATURE,
    "translation": _TEXT2TEXT_SIGNATURE,
    "text-generation": _TEXT2TEXT_SIGNATURE,
    "text2text-generation": _TEXT2TEXT_SIGNATURE,
    "text-classification": _CLASSIFICATION_SIGNATURE,
    "conversational": _TEXT2TEXT_SIGNATURE,
    "fill-mask": _TEXT2TEXT_SIGNATURE,
    "summarization": _TEXT2TEXT_SIGNATURE,
    "image-classification": _CLASSIFICATION_SIGNATURE,
    "zero-shot-classification": ModelSignature(
        inputs=Schema(
            [
                ColSpec(DataType.string, name="sequences"),
                ColSpec(DataType.string, name="candidate_labels"),
                ColSpec(DataType.string, name="hypothesis_template"),
            ]
        ),
        outputs=Schema(
            [
                ColSpec(DataType.string, name="sequence"),
                ColSpec(DataType.string, name="labels"),
                ColSpec(DataType.double, name="scores"),
            ]
        ),
    ),
    "automatic-speech-recognition": ModelSignature(
        inputs=Schema([ColSpec(DataType.binary)]),
        outputs=Schema([ColSpec(DataType.string)]),
    ),
    "audio-classification": ModelSignature(
        inputs=Schema([ColSpec(DataType.binary)]),
        outputs=Schema(
            [ColSpec(DataType.double, name="score"), ColSpec(DataType.string, name="label")]
        ),
    ),
    "table-question-answering": ModelSignature(
        inputs=Schema(
            [ColSpec(DataType.string, name="query"), ColSpec(DataType.string, name="table")]
        ),
        outputs=Schema([ColSpec(DataType.string)]),
    ),
    "question-answering": ModelSignature(
        inputs=Schema(
            [ColSpec(DataType.string, name="question"), ColSpec(DataType.string, name="context")]
        ),
        outputs=Schema([ColSpec(DataType.string)]),
    ),
    "feature-extraction": ModelSignature(
        inputs=Schema([ColSpec(DataType.string)]),
        outputs=Schema([TensorSpec(np.dtype("float64"), [-1], "double")]),
    ),
}


def infer_or_get_default_signature(
    pipeline, example=None, model_config=None, flavor_config=None
) -> ModelSignature:
    """
    Assigns a default ModelSignature for a given Pipeline type that has pyfunc support. These
    default signatures should only be generated and assigned when saving a model iff the user
    has not supplied a signature.
    For signature inference in some Pipelines that support complex input types, an input example
    is needed.
    """
    import transformers

    if example is not None and isinstance(pipeline, transformers.Pipeline):
        try:
            timeout = MLFLOW_INPUT_EXAMPLE_INFERENCE_TIMEOUT.get()
            if timeout and is_windows():
                timeout = None
                _logger.warning(
                    "On Windows, timeout is not supported for model signature inference. "
                    "Therefore, the operation is not bound by a timeout and may hang indefinitely. "
                    "If it hangs, please consider specifying the signature manually."
                )
            return _infer_signature_with_example(
                pipeline, example, model_config, flavor_config, timeout
            )
        except Exception as e:
            if isinstance(e, MlflowTimeoutError):
                msg = (
                    "Attempted to generate a signature for the saved pipeline but prediction timed "
                    f"out after {timeout} seconds. Falling back to the default signature for the "
                    "pipeline. You can specify a signature manually or increase the timeout "
                    f"by setting the environment variable {MLFLOW_INPUT_EXAMPLE_INFERENCE_TIMEOUT}"
                )
            else:
                msg = (
                    "Attempted to generate a signature for the saved pipeline but encountered an "
                    f"error. Fall back to the default signature for the pipeline type. Error: {e}"
                )
            _logger.warning(msg)

    task = getattr(pipeline, "task", None)
    if task.startswith("translation_"):
        task = "translation"
    if signature := _DEFAULT_SIGNATURE_FOR_TASK.get(task):
        return signature

    _logger.warning(
        "An unsupported task type was supplied for signature inference. Either provide an "
        "`input_example` or generate a signature manually via `infer_signature` to have a "
        "signature recorded in the MLmodel file."
    )


def _infer_signature_with_example(
    pipeline, example, model_config=None, flavor_config=None, timeout=None
) -> ModelSignature:
    params = None
    if _contains_params(example):
        example, params = example
    example = format_input_example_for_special_cases(example, pipeline)

    if timeout:
        _logger.info(
            "Running model prediction to infer the model output signature with a timeout "
            f"of {timeout} seconds. You can specify a different timeout by setting the "
            f"environment variable {MLFLOW_INPUT_EXAMPLE_INFERENCE_TIMEOUT}."
        )
        with run_with_timeout(timeout):
            prediction = generate_signature_output(
                pipeline, example, model_config, flavor_config, params
            )
    else:
        prediction = generate_signature_output(
            pipeline, example, model_config, flavor_config, params
        )
    return infer_signature(example, prediction, params)


def format_input_example_for_special_cases(input_example, pipeline):
    """
    Handles special formatting for specific types of Pipelines so that the displayed example
    reflects the correct example input structure that mirrors the behavior of the input parsing
    for pyfunc.
    """
    input_data = input_example[0] if isinstance(input_example, tuple) else input_example

    if (
        pipeline.task == "zero-shot-classification"
        and isinstance(input_data, dict)
        and isinstance(input_data["candidate_labels"], list)
    ):
        input_data["candidate_labels"] = json.dumps(input_data["candidate_labels"])
    return input_data if not isinstance(input_example, tuple) else (input_data, input_example[1])


def generate_signature_output(pipeline, data, model_config=None, flavor_config=None, params=None):
    # Lazy import to avoid circular dependencies. Ideally we should move _TransformersWrapper
    # out from __init__.py to avoid this.
    from mlflow.transformers import _TransformersWrapper

    return _TransformersWrapper(
        pipeline=pipeline, model_config=model_config, flavor_config=flavor_config
    ).predict(data, params=params)



from typing import Any

from pydantic import ConfigDict

from mlflow.types.chat import BaseModel, ChatUsage, ToolCall
from mlflow.types.llm import (
    _custom_inputs_col_spec,
    _custom_outputs_col_spec,
    _token_usage_stats_col_spec,
)
from mlflow.types.schema import (
    Array,
    ColSpec,
    DataType,
    Map,
    Object,
    Property,
    Schema,
)
from mlflow.utils.pydantic_utils import IS_PYDANTIC_V2_OR_NEWER, model_validator


class ChatAgentMessage(BaseModel):
    """
    A message in a ChatAgent model request or response.

    Args:
        role (str): The role of the entity that sent the message (e.g. ``"user"``, ``"system"``,
            ``"assistant"``, ``"tool"``).
        content (str): The content of the message.
            **Optional** Can be ``None`` if tool_calls is provided.
        name (str): The name of the entity that sent the message. **Optional** defaults to ``None``
        id (str): The ID of the message. Required when it is either part of a
            :py:class:`ChatAgentResponse` or :py:class:`ChatAgentChunk`.
        tool_calls (List[:py:class:`mlflow.types.chat.ToolCall`]): A list of tool calls made by the
            model. **Optional** defaults to ``None``
        tool_call_id (str): The ID of the tool call that this message is a response to.
            **Optional** defaults to ``None``
        attachments (Dict[str, str]): A dictionary of attachments. **Optional** defaults to ``None``
    """

    role: str
    content: str | None = None
    name: str | None = None
    id: str | None = None
    tool_calls: list[ToolCall] | None = None
    tool_call_id: str | None = None
    # TODO make this a pydantic class with subtypes once we have more details on usage
    attachments: dict[str, str] | None = None

    @model_validator(mode="after")
    def check_content_and_tool_calls(cls, values):
        """
        Ensure at least one of 'content' or 'tool_calls' is set.
        """
        if IS_PYDANTIC_V2_OR_NEWER:
            content = values.content
            tool_calls = values.tool_calls
        else:
            content = values.get("content")
            tool_calls = values.get("tool_calls")

        if content is None and tool_calls is None:
            raise ValueError("Either 'content' or 'tool_calls' must be provided.")
        return values

    @model_validator(mode="after")
    def check_tool_messages(cls, values):
        """
        Ensure that the 'name' and 'tool_call_id' fields are set for tool messages.
        """
        if IS_PYDANTIC_V2_OR_NEWER:
            name = values.name
            role = values.role
            tool_call_id = values.tool_call_id
        else:
            name = values.get("name")
            role = values.get("role")
            tool_call_id = values.get("tool_call_id")

        if role == "tool" and (not name or not tool_call_id):
            raise ValueError("Both 'name' and 'tool_call_id' must be provided for tool messages.")
        return values


class ChatContext(BaseModel):
    """
    Context to be used in a ChatAgent endpoint.

    Args:
        conversation_id (str): The ID of the conversation. **Optional** defaults to ``None``
        user_id (str): The ID of the user. **Optional** defaults to ``None``
    """

    conversation_id: str | None = None
    user_id: str | None = None


class ChatAgentRequest(BaseModel):
    """
    Format of a ChatAgent interface request.

    Args:
        messages: A list of :py:class:`ChatAgentMessage` that will be passed to the model.
        context (:py:class:`ChatContext`): The context to be used in the chat endpoint. Includes
            conversation_id and user_id. **Optional** defaults to ``None``
        custom_inputs (Dict[str, Any]): An optional param to provide arbitrary additional context
            to the model. The dictionary values must be JSON-serializable.
            **Optional** defaults to ``None``
        stream (bool): Whether to stream back responses as they are generated.
            **Optional**, defaults to ``False``
    """

    messages: list[ChatAgentMessage]
    context: ChatContext | None = None
    custom_inputs: dict[str, Any] | None = None
    stream: bool | None = False


class ChatAgentResponse(BaseModel):
    """
    Represents the response of a ChatAgent.

    Args:
        messages: A list of :py:class:`ChatAgentMessage` that are returned from the model.
        finish_reason (str): The reason why generation stopped. **Optional** defaults to ``None``
        custom_outputs (Dict[str, Any]): An optional param to provide arbitrary additional context
            from the model. The dictionary values must be JSON-serializable. **Optional**, defaults
            to ``None``
        usage (:py:class:`mlflow.types.chat.ChatUsage`): The token usage of the request
            **Optional**, defaults to None
    """

    if IS_PYDANTIC_V2_OR_NEWER:
        model_config = ConfigDict(validate_assignment=True)
    else:

        class Config:
            validate_assignment = True

    messages: list[ChatAgentMessage]
    finish_reason: str | None = None
    # TODO: add finish_reason_metadata once we have a plan for usage
    custom_outputs: dict[str, Any] | None = None
    usage: ChatUsage | None = None

    @model_validator(mode="after")
    def check_message_ids(cls, values):
        """
        Ensure that all messages have an ID and it is unique.
        """
        if IS_PYDANTIC_V2_OR_NEWER:
            message_ids = [msg.id for msg in values.messages]
        else:
            message_ids = [msg.get("id") for msg in values.get("messages")]

        if any(msg_id is None for msg_id in message_ids):
            raise ValueError(
                "All ChatAgentMessage objects in field `messages` must have an ID. You can use "
                "`str(uuid.uuid4())` to generate a unique ID."
            )
        if len(message_ids) != len(set(message_ids)):
            raise ValueError(
                "All ChatAgentMessage objects in field `messages` must have unique IDs. "
                "You can use `str(uuid.uuid4())` to generate a unique ID."
            )
        return values


class ChatAgentChunk(BaseModel):
    """
    Represents a single chunk within the streaming response of a ChatAgent.

    Args:
        delta: A :py:class:`ChatAgentMessage` representing a single chunk within the list of
            messages comprising agent output. In particular, clients should assume the `content`
            field within this `ChatAgentMessage` contains only part of the message content, and
            aggregate message content by ID across chunks. More info can be found in the docstring
            of :py:func:`ChatAgent.predict_stream <mlflow.pyfunc.ChatAgent.predict_stream>`.
        finish_reason (str): The reason why generation stopped. **Optional** defaults to ``None``
        custom_outputs (Dict[str, Any]): An optional param to provide arbitrary additional context
            from the model. The dictionary values must be JSON-serializable. **Optional**, defaults
            to ``None``
        usage (:py:class:`mlflow.types.chat.ChatUsage`): The token usage of the request
            **Optional**, defaults to None
    """

    if IS_PYDANTIC_V2_OR_NEWER:
        model_config = ConfigDict(validate_assignment=True)
    else:

        class Config:
            validate_assignment = True

    delta: ChatAgentMessage
    finish_reason: str | None = None
    # TODO: add finish_reason_metadata once we have a plan for usage
    custom_outputs: dict[str, Any] | None = None
    usage: ChatUsage | None = None

    @model_validator(mode="after")
    def check_message_id(cls, values):
        """
        Ensure that the message ID is unique.
        """
        message_id = values.delta.id if IS_PYDANTIC_V2_OR_NEWER else values.get("delta").get("id")

        if message_id is None:
            raise ValueError(
                "The field `delta` of ChatAgentChunk must contain a ChatAgentMessage object with an"
                " ID. If this chunk contains partial content, it should have the same ID as other "
                " chunks in the same message. See "
                "https://mlflow.org/docs/latest/api_reference/python_api/mlflow.pyfunc.html#mlflow.pyfunc.ChatAgent.predict_stream"
                " for more details. You can use `str(uuid.uuid4())` to generate a unique ID."
            )
        return values


# fmt: off
_chat_agent_messages_col_spec = ColSpec(
    name="messages",
    type=Array(
        Object(
            [
                Property("role", DataType.string),
                Property("content", DataType.string, False),
                Property("name", DataType.string, False),
                Property("id", DataType.string, False),
                Property("tool_calls", Array(Object([
                    Property("id", DataType.string),
                    Property("function", Object([
                        Property("name", DataType.string),
                        Property("arguments", DataType.string),
                    ])),
                    Property("type", DataType.string),
                ])), False),
                Property("tool_call_id", DataType.string, False),
                Property("attachments", Map(DataType.string), False),
            ]
        )
    ),
)

# TODO: move out all params to a ParamSchema when Map(AnyType()) is supported by ParamSpec
CHAT_AGENT_INPUT_SCHEMA = Schema(
    [
        _chat_agent_messages_col_spec,
        ColSpec(name="context", type=Object([
            Property("conversation_id", DataType.string, False),
            Property("user_id", DataType.string, False),
        ]), required=False),
        _custom_inputs_col_spec,
        ColSpec(name="stream", type=DataType.boolean, required=False),
    ]
)

CHAT_AGENT_OUTPUT_SCHEMA = Schema(
    [
        _chat_agent_messages_col_spec,
        ColSpec(name="finish_reason", type=DataType.string, required=False),
        _custom_outputs_col_spec,
        _token_usage_stats_col_spec,
    ]
)

CHAT_AGENT_INPUT_EXAMPLE = {
    "messages": [
        {"role": "user", "content": "Hello!"},
    ]
}



"""
Introduces main Context class and the framework to specify different specialized
contexts.
"""

import functools
from abc import ABC, abstractmethod
from typing import Callable, ParamSpec, TypeVar

import mlflow
from mlflow.tracking.context import registry as context_registry
from mlflow.utils.mlflow_tags import MLFLOW_USER

P = ParamSpec("P")
R = TypeVar("R")


class Context(ABC):
    """
    Abstract class for execution context.
    Context is stateless and should NOT be used to store information related to specific eval run.
    """

    @abstractmethod
    def get_mlflow_experiment_id(self) -> str | None:
        """
        Get the current MLflow experiment ID, or None if not running within an MLflow experiment.
        """

    @abstractmethod
    def get_mlflow_run_id(self) -> str | None:
        """
        Gets the MLflow RunId, or None if not running within an MLflow run.
        """

    @abstractmethod
    def get_user_name(self) -> str:
        """
        Get the current user's name.
        """


class NoneContext(Context):
    """
    A context that does nothing.
    """

    def get_mlflow_experiment_id(self) -> str | None:
        raise NotImplementedError("Context is not set")

    def get_mlflow_run_id(self) -> str | None:
        raise NotImplementedError("Context is not set")

    def get_user_name(self) -> str:
        raise NotImplementedError("Context is not set")


class RealContext(Context):
    """
    Context for eval execution.

    NOTE: This class is not covered by unit tests and is meant to be tested through
    smoke tests that run this code on an actual Databricks cluster.
    """

    def __init__(self):
        self._run_id = None
        self._context_tags = context_registry.resolve_tags()

    def get_mlflow_experiment_id(self) -> str | None:
        # Note `_get_experiment_id` is thread-safe
        return mlflow.tracking.fluent._get_experiment_id()

    def get_mlflow_run_id(self) -> str | None:
        """
        Gets the MLflow run_id the evaluation harness is running under.

        Warning: This run_id may not be active. This happens when `get_mlflow_run_id` is called from
        a different thread than the one that started the MLflow run.
        """
        # First check if a run ID is specified explicitly by the parent thread
        if self._run_id:
            return self._run_id

        # Otherwise fall back to the active run in the current thread
        if run := mlflow.active_run():
            return run.info.run_id

        return None

    def set_mlflow_run_id(self, run_id: str) -> None:
        """
        Set the MLflow run ID explicitly.

        This method should be called when running code in a different thread than the one that
        started the MLflow run. It sets the run ID in a thread-local variable so that it can be
        accessed from the thread.
        """
        self._run_id = run_id

    def get_user_name(self) -> str:
        return self._context_tags.get(MLFLOW_USER, "unknown")


# Context is a singleton.
_context_singleton = NoneContext()


def context_is_active() -> bool:
    """
    Check if a context is active.
    """
    return not isinstance(get_context(), NoneContext)


def get_context() -> Context:
    """
    Get the context.
    """
    return _context_singleton


def eval_context(func: Callable[P, R]) -> Callable[P, R]:
    """
    Decorator for wrapping all eval APIs with setup and closure logic.

    Sets up a context singleton with RealContext if there isn't one already.
    """

    @functools.wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
        # Set up the context singleton if it doesn't exist
        if not context_is_active():
            global _context_singleton
            _context_singleton = RealContext()

        return func(*args, **kwargs)

    return wrapper


def _set_context(context: Context) -> None:
    """SHOULD ONLY BE USED FOR TESTING."""
    global _context_singleton
    _context_singleton = context



from __future__ import annotations

from typing import TYPE_CHECKING

from mlflow.exceptions import MlflowException
from mlflow.protos.databricks_pb2 import INVALID_PARAMETER_VALUE

if TYPE_CHECKING:
    import torch

_TORCH_DTYPE_KEY = "torch_dtype"


def _extract_torch_dtype_if_set(pipeline) -> torch.dtype | None:
    """
    Extract the torch datatype argument if set and return as a string encoded value.
    """
    try:
        import torch
    except ImportError:
        # If torch is not installed, safe to assume the model doesn't have a custom torch_dtype
        return None

    # Check model dtype as pipeline's torch_dtype field doesn't always reflect the model's dtype
    model_dtype = pipeline.model.dtype if hasattr(pipeline.model, "dtype") else None

    # If the underlying model is PyTorch model, dtype must be a torch.dtype instance
    return model_dtype if isinstance(model_dtype, torch.dtype) else None


def _deserialize_torch_dtype(dtype_str: str) -> torch.dtype:
    """
    Convert the string-encoded `torch_dtype` pipeline argument back to the correct `torch.dtype`
    instance value for applying to a loaded pipeline instance.
    """
    try:
        import torch
    except ImportError as e:
        raise MlflowException(
            "Unable to determine if the value supplied by the argument "
            "torch_dtype is valid since torch is not installed.",
            error_code=INVALID_PARAMETER_VALUE,
        ) from e

    if dtype_str.startswith("torch."):
        dtype_str = dtype_str[6:]

    dtype = getattr(torch, dtype_str, None)
    if isinstance(dtype, torch.dtype):
        return dtype

    raise MlflowException(
        f"The value '{dtype_str}' is not a valid torch.dtype",
        error_code=INVALID_PARAMETER_VALUE,
    )



import os
import posixpath
import sys
import urllib.parse
from contextlib import contextmanager
from queue import Queue

from mlflow.entities import FileInfo
from mlflow.store.artifact.artifact_repo import ArtifactRepository


# Based on: https://stackoverflow.com/a/58466685
def _put_r_for_windows(sftp, local_dir, remote_dir, preserve_mtime=False):
    for entry in os.listdir(local_dir):
        local_path = os.path.join(local_dir, entry)
        remote_path = posixpath.join(remote_dir, entry)
        if os.path.isdir(local_path):
            sftp.mkdir(remote_path)
            _put_r_for_windows(sftp, local_path, remote_path, preserve_mtime)
        else:
            sftp.put(local_path, remote_path, preserve_mtime=preserve_mtime)


class _SftpPool:
    def __init__(self, connections):
        self._idle_connection_queue = Queue()
        for c in connections:
            self._idle_connection_queue.put(c)

    @contextmanager
    def get_sftp_connection(self):
        c = self._idle_connection_queue.get(block=True)
        try:
            yield c
        finally:
            self._idle_connection_queue.put(c)


class SFTPArtifactRepository(ArtifactRepository):
    """Stores artifacts as files in a remote directory, via sftp."""

    def __init__(
        self, artifact_uri: str, tracking_uri: str | None = None, registry_uri: str | None = None
    ) -> None:
        super().__init__(artifact_uri, tracking_uri, registry_uri)
        parsed = urllib.parse.urlparse(artifact_uri)
        self.config = {
            "host": parsed.hostname,
            "port": parsed.port,
            "username": parsed.username,
            "password": parsed.password,
        }
        self.path = parsed.path or "/"

        import paramiko
        import pysftp

        if self.config["host"] is None:
            self.config["host"] = "localhost"

        ssh_config = paramiko.SSHConfig()
        user_config_file = os.path.expanduser("~/.ssh/config")
        if os.path.exists(user_config_file):
            with open(user_config_file) as f:
                ssh_config.parse(f)

        user_config = ssh_config.lookup(self.config["host"])

        if "hostname" in user_config:
            self.config["host"] = user_config["hostname"]

        if self.config.get("username", None) is None and "user" in user_config:
            self.config["username"] = user_config["user"]

        if self.config.get("port", None) is None:
            if "port" in user_config:
                self.config["port"] = int(user_config["port"])
            else:
                self.config["port"] = 22

        if "identityfile" in user_config:
            self.config["private_key"] = user_config["identityfile"][0]

        connections = [pysftp.Connection(**self.config) for _ in range(self.max_workers)]
        self.pool = _SftpPool(connections)

    def log_artifact(self, local_file, artifact_path=None):
        artifact_dir = posixpath.join(self.path, artifact_path) if artifact_path else self.path
        with self.pool.get_sftp_connection() as sftp:
            sftp.makedirs(artifact_dir)
            sftp.put(local_file, posixpath.join(artifact_dir, os.path.basename(local_file)))

    def log_artifacts(self, local_dir, artifact_path=None):
        artifact_dir = posixpath.join(self.path, artifact_path) if artifact_path else self.path
        with self.pool.get_sftp_connection() as sftp:
            sftp.makedirs(artifact_dir)
            if sys.platform == "win32":
                _put_r_for_windows(sftp, local_dir, artifact_dir)
            else:
                sftp.put_r(local_dir, artifact_dir)

    def _is_directory(self, artifact_path):
        artifact_dir = self.path
        path = posixpath.join(artifact_dir, artifact_path) if artifact_path else artifact_dir
        with self.pool.get_sftp_connection() as sftp:
            return sftp.isdir(path)

    def list_artifacts(self, path=None):
        artifact_dir = self.path
        list_dir = posixpath.join(artifact_dir, path) if path else artifact_dir
        with self.pool.get_sftp_connection() as sftp:
            if not sftp.isdir(list_dir):
                return []
            artifact_files = sftp.listdir(list_dir)
            infos = []
            for file_name in artifact_files:
                file_path = file_name if path is None else posixpath.join(path, file_name)
                full_file_path = posixpath.join(list_dir, file_name)
                if sftp.isdir(full_file_path):
                    infos.append(FileInfo(file_path, True, None))
                else:
                    infos.append(FileInfo(file_path, False, sftp.stat(full_file_path).st_size))
            return infos

    def _download_file(self, remote_file_path, local_path):
        remote_full_path = posixpath.join(self.path, remote_file_path)
        with self.pool.get_sftp_connection() as sftp:
            sftp.get(remote_full_path, local_path)

    def delete_artifacts(self, artifact_path=None):
        artifact_dir = posixpath.join(self.path, artifact_path) if artifact_path else self.path
        with self.pool.get_sftp_connection() as sftp:
            self._delete_inner(artifact_dir, sftp)

    def _delete_inner(self, artifact_path, sftp):
        if sftp.isdir(artifact_path):
            with sftp.cd(artifact_path):
                for element in sftp.listdir():
                    self._delete_inner(element, sftp)
            sftp.rmdir(artifact_path)
        elif sftp.isfile(artifact_path):
            sftp.remove(artifact_path)
